#
# CVODES-driven single cell simulation with optional sensitivities and root
# finding
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import os
import platform
import tempfile

from collections import OrderedDict

import myokit

# Location of C template
SOURCE_FILE = 'cvodessim.c'


class Simulation(myokit.CModule):
    """
    Runs single cell simulations using the CVODES solver (see [1]); CVODES uses
    an implicit multi-step method to achieve high accuracy and stability with
    adaptive step sizes.

    The model passed to the simulation is cloned and stored internally, so
    changes to the original model object will not affect the simulation. A
    protocol can be passed in as ``protocol`` or set later using
    :meth:`set_protocol`.

    Simulations maintain an internal state consisting of

    - the current simulation time
    - the current state
    - the default state
    - (optional) the current and default state's sensitivities with respect to
      the selected independent variables.

    When a simulation is created, the simulation time is set to 0 and both the
    current and the default state are copied from the model.
    After each call to :meth:`Simulation.run` the time variable and current
    state are updated, so that each successive call to run continues where the
    previous simulation left off. A :meth:`reset` method is provided that will
    set the time back to 0 and revert the current state to the default state.
    To change the time or state manually, use :meth:`set_time` and
    :meth:`set_state`.

    A pre-pacing method :meth:`pre` is provided that doesn't affect the
    simulation time but will update the current *and the default state*. This
    allows you to pre-pace, run a simulation, reset to the pre-paced state, run
    another simulation etc.

    **Sensitivities**

    Sensitivities of model variables with respect to parameters (any unbound
    model variable defined as a literal constant) or initial values of states
    can be calculated using the ``sensitivities`` argument. If set, this should
    be a tuple ``(ys, xs)`` containing the "dependent" and "independent"
    expressions used to make up the calculated sensitivities (partial
    derivatives) ``dy/dx``.

    The "dependent variables" must refer to state variables, time-derivatives
    of state variables, or intermediary variables. These can be specified as
    :class:`myokit.Variable` objects, as :class:`myokit.Name` or
    :class:`myokit.Derivative` expressions, or as strings e.g. ``"ina.INa"`` or
    ``"dot(membrane.V)"``.

    The "independent variables" in sensitivities must refer to either literal
    variables (variables with no dependencies) or initial values of state
    variables. These can be specified as :class:`myokit.Variable` objects, as
    :class:`myokit.Name` or :class:`myokit.InitialValue` expressions, or as
    strings e.g. ``"ikr.gKr"`` or ``"init(membrane.V)"``.

    **Bound variables and labels**

    The simulation provides four inputs a model variable can be bound to:

    ``time``
        This input provides the simulation time.
    ``pace``
        This input provides the current value of the pacing variable. This is
        determined using the protocol passed into the Simulation.
    ``evaluations``
        This input provides the number of rhs evaluations used at each point in
        time and can be used to gain some insight into the solver's behavior.
    ``realtime``
        This input provides the elapsed system time at each logged point.

    No variable labels are required for this simulation type.

    **Multiple protocols or no protocol**

    This simulation supports pacing with more than one protocol. To this end,
    pass in a dictionary mapping pacing labels (bindings) to :class:`Protocol`
    or :class:`TimeSeriesProtocol` objects, e.g.
    ``protocol={'pace_1': protocol_1, 'pace_2': protocol_2}``.

    For backwards compatibility, if no protocol is set and ``protocol=None``,
    then the pacing label ``pace`` is still registered (allowing later calls to
    add a protocol with ``set_protocol``. Alternatively, if ``protocol={}``
    then no pacing labels will be registered, and any subsequent calls to
    ``set_protocol`` will fail.

    **Storing and loading simulation objects**

    There are two ways to store Simulation objects to the file system: 1.
    using serialisation (the ``pickle`` module), and 2. using the ``path``
    constructor argument.

    Each time a simulation is created, a C module is compiled, imported, and
    deleted in the background. This means that part of a ``Simulation`` object
    is a reference to an imported C extension, and cannot be serialized. When
    using ``pickle`` to serialize the simulation, this compiled part is not
    stored. Instead, when the pickled simulation is read from disk the
    compilation is repeated. Unpickling simulations also restores their state:
    variables such as model state, simulation time, and tolerance are preserved
    when pickling.

    As an alternative to serialisation, Simulations can be created with a
    ``path`` variable that specifies a location where the generated module can
    be stored. For example, calling ``Simulation(model, path='my-sim.zip')``
    will create a file called ``my-sim.zip`` in which the C extension is
    stored. To load, use ``Simulation.from_path('my-sim.zip')``. Unlike
    pickling, simulations stored and loaded this way do not maintain state:
    variables such as model state, simulation time, and tolerance are not
    preserved with this method. Finally, note that (again unlike pickled
    simulations), the generated zip files are highly platform dependent: a zip
    file generated on one machine may not work on another.

    **Arguments**

    ``model``
        The model to simulate
    ``protocol``
        A :class:`myokit.Protocol` or :class:`myokit.TimeSeriesProtocol` to use
        for the variable with binding ``pace``. Atlernatively, a dictionary
        mapping binding labels to :class:`myokit.Protocol` objects can be used
        to run with multiple protocols. Finally, can be ``None`` to run without
        a protocol.
    ``sensitivities``
        An optional tuple ``(dependents, independents)`` where ``dependents``
        is a list of variables or expressions to take derivatives of (``y`` in
        ``dy/dx``) and ``independents`` is a list of variables or expressions
        to calculate derivatives to (``x`` in ``dy/dx``).
        Each entry in ``dependents`` must be a :class:`myokit.Variable`, a
        :class:`myokit.Name`, a :class:`myokit.Derivative`, or a string with
        either a fully qualified variable name or a ``dot()`` expression.
        Each entry in ``independents`` must be a :class:`myokit.Variable`, a
        :class:`myokit.Name`, a :class:`myokit.InitialValue`, or a string with
        either a fully qualified variable name or an ``init()`` expression.
    ``path``
        An optional path used to load or store compiled simulation objects. See
        "Storing and loading simulation objects", above.

    **References**

    [1] SUNDIALS: Suite of nonlinear and differential/algebraic equation
    solvers. Hindmarsh, Brown, Woodward, et al. (2005) ACM Transactions on
    Mathematical Software.

    """
    _index = 0  # Simulation id

    def __init__(self, model, protocol=None, sensitivities=None, path=None):
        super().__init__()

        # Require a valid model
        if not model.is_valid():
            model.validate()
        self._model = model.clone()
        del model

        # Set protocol
        self._protocols = []
        self._pacing_labels = []
        if isinstance(protocol, (myokit.Protocol, myokit.TimeSeriesProtocol)):
            protocol = {'pace': protocol}
        elif protocol is None:
            # For backwards compatibility, we still register 'pace'. This
            # means users can call `set_protocol` at a later time to set a
            # protocol.
            protocol = {'pace': None}
        for label, protocol in protocol.items():
            self._protocols.append(None)
            self._pacing_labels.append(label)
            self.set_protocol(protocol, label)

        # Generate C Model code, get sensitivity and constants info
        cmodel = myokit.CModel(self._model, self._pacing_labels, sensitivities)
        if cmodel.has_sensitivities:
            self._sensitivities = (cmodel.dependents, cmodel.independents)

            # Check for sensitivities w.r.t. variables used in initial state
            # expressions. This is not implemented yet.
            inits = self._model.initial_values()
            for i in self._sensitivities[1]:  # Expressions
                if isinstance(i, myokit.Name):
                    for e in inits:
                        if e.depends_on(i, deep=True):
                            raise NotImplementedError(
                                'Sensitivities with respect to parameters used'
                                ' in initial conditions is not implemented ('
                                + e.code() + ' depends on ' + i.code() + ').')
        else:
            self._sensitivities = None

        # Ordered dicts mapping Variable objects to float values
        self._literals = OrderedDict()
        self._parameters = OrderedDict()
        for var, eq in cmodel.literals.items():
            self._literals[var] = eq.rhs.eval()
        for var, eq in cmodel.parameters.items():
            self._parameters[var] = eq.rhs.eval()

        # Compile or load simulation
        if type(path) == tuple:
            path, self._sim = path
        else:
            self._create_simulation(cmodel.code, path)
        del cmodel

        # Get state and default state from model
        self._state = self._model.initial_values(as_floats=True)
        self._default_state = list(self._state)

        # Set state and default state for sensitivities
        self._s_state = self._s_default_state = None
        if self._sensitivities:
            # Outer index: number of independent variables
            # Inner index: number of states
            self._s_state = []
            for expr in self._sensitivities[1]:
                row = [0.0] * len(self._state)
                if isinstance(expr, myokit.InitialValue):
                    row[expr.var().index()] = 1.0
                self._s_state.append(row)
            self._s_default_state = [list(x) for x in self._s_state]

        # Last state reached before error
        self._error_state = None

        # Starting time
        self._time = 0

        # Set default min and max step size
        self._dtmax = self._dtmin = None

        # Set default tolerance values
        self._tolerance = None
        self.set_tolerance()

    def _create_simulation(self, cmodel_code, path):
        """
        Creates and compiles the C simulation module.

        If ``path`` is set, also stores the built extension in a zip file at
        ``path``, along with a serialisation of the Python Simulation object.
        """
        # Unique simulation id
        Simulation._index += 1
        module_name = 'myokit_sim_' + str(Simulation._index)
        module_name += '_' + str(myokit.pid_hash())

        # Arguments
        args = {
            'module_name': module_name,
            'model_code': cmodel_code,
        }
        fname = os.path.join(myokit.DIR_CFUNC, SOURCE_FILE)

        # Define libraries
        libs = [
            'sundials_cvodes',
            'sundials_nvecserial',
        ]
        if platform.system() != 'Windows':  # pragma: no windows cover
            libs.append('m')

        # Define library paths
        # Note: Sundials path on windows already includes local binaries
        libd = list(myokit.SUNDIALS_LIB)
        incd = list(myokit.SUNDIALS_INC)
        incd.append(myokit.DIR_CFUNC)

        # Create extension
        store = path is not None
        res = self._compile(
            module_name, fname, args, libs, libd, incd, store_build=store)

        # Store module and return
        if store:
            self._sim, d_build = res
            self._store_build(path, d_build, module_name)
        else:
            self._sim = res

    def _store_build(self, path, d_build, name):
        """
        Stores this simulation to ``path``, including all information from the
        build directory ``d_build`` and the generated backend module's
        ``name``.
        """
        try:
            import zipfile
            try:
                import zlib     # noqa
            except ImportError:
                raise Exception('Storing simulations requires the zlib module'
                                ' to be available.')

            # Sensitivity constructor argument. Don't pass in expressions, as
            # they will need to pickle the variables as well, which in turn
            # will need a component, model, etc.
            sens_arg = None
            if self._sensitivities:
                sens_arg = [[x.code() for x in y] for y in self._sensitivities]

            # Store serialized name and constructor args
            fname = os.path.join(d_build, '_simulation.pickle')
            import pickle
            with open(fname, 'wb') as f:
                pickle.dump(str(name), f)
                pickle.dump(self._model, f)
                pickle.dump(self._protocols, f)
                pickle.dump(self._pacing_labels, f)
                pickle.dump(sens_arg, f)

            # Zip it all in
            with zipfile.ZipFile(path, 'w', zipfile.ZIP_DEFLATED) as f:
                for root, dirs, files in os.walk(d_build):
                    for fname in files:
                        org = os.path.join(root, fname)
                        new = os.path.relpath(org, d_build)
                        f.write(org, new)
        finally:
            myokit.tools.rmtree(d_build, silent=True)

    @staticmethod
    def from_path(path):
        """
        Loads a simulation from the zip file in ``path``.

        The file at ``path`` must have been created by :class:`Simulation` and
        on the same platform (e.g. the same operating system, processor
        architecture, Myokit version etc.).

        Note that the simulation state (e.g. time, model state, solver
        tolerance etc. is not restored).

        Returns a :class:`Simulation` object.
        """
        import zipfile
        try:
            import zlib     # noqa
        except ImportError:
            raise Exception('Loading simulations requires the zlib module to'
                            ' be available.')

        # Unpack the zipped information
        d_build = tempfile.mkdtemp('myokit_build')
        try:
            with zipfile.ZipFile(path, 'r', zipfile.ZIP_DEFLATED) as f:
                f.extractall(d_build)

            # Load serialized name and constructor args
            fname = os.path.join(d_build, '_simulation.pickle')
            import pickle
            with open(fname, 'rb') as f:
                name = pickle.load(f)
                model = pickle.load(f)
                protocols = pickle.load(f)
                pacing_labels = pickle.load(f)
                sensitivities = pickle.load(f)

            # Load module
            from myokit._sim import load_module
            module = load_module(name, d_build)

            # Create and return simulation
            labeled_protocols = {
                k: v for k, v in zip(pacing_labels, protocols)
            }
            return Simulation(
                model, labeled_protocols, sensitivities, (path, module)
            )

        finally:
            myokit.tools.rmtree(d_build, silent=True)

    def default_state(self):
        """
        Returns the default state.
        """
        return list(self._default_state)

    def default_state_sensitivities(self):
        """
        Returns the default sensitivities with respect to state variables, or
        ``None`` if sensitivities are not enabled.
        """
        if self._sensitivities:
            return [list(x) for x in self._s_default_state]
        return None

    def last_state(self):
        """
        If the last call to :meth:`Simulation.pre()` or
        :meth:`Simulation.run()` resulted in an error, this will return the
        last state reached during that simulation.

        Will return ``None`` if no simulation was run or the simulation did not
        result in an error.
        """
        return list(self._error_state) if self._error_state else None

    def eval_derivatives(self, y=None, pacing=None):
        """
        Evaluates and returns the state derivatives.

        The state to evaluate for can be given as ``y``. If no state is given
        the current simulation state is used.

        An optional dict ``pacing`` can be passed in that maps labels to
        numerical values. If a label is used but not provided in ``pacing``,
        it's value will be set to 0.
        """
        # Get state
        if y is None:
            y = list(self._state)
        else:
            y = self._model.map_to_state(y)

        pacing_values = [0.0] * len(self._pacing_labels)
        if pacing is not None:
            for k, v in pacing.items():
                try:
                    ki = self._pacing_labels.index(k)
                    pacing_values[ki] = float(v)
                except ValueError:
                    pass

        # Create space to store derivatives
        dy = list(self._state)

        # Evaluate and return
        self._sim.eval_derivatives(
            # 0. Time
            0,
            # 1. Pace
            pacing_values,
            # 2. State
            y,
            # 3. Space to store the state derivatives
            dy,
            # 4. Literal values
            list(self._literals.values()),
            # 5. Parameter values
            list(self._parameters.values()),
        )
        return dy

    def last_number_of_evaluations(self):
        """
        Returns the number of rhs evaluations performed by the solver during
        the last simulation.
        """
        return self._sim.number_of_evaluations()

    def last_number_of_steps(self):
        """
        Returns the number of steps taken by the solver during the last
        simulation.
        """
        return self._sim.number_of_steps()

    def pre(self, duration, progress=None, msg='Pre-pacing simulation'):
        """
        This method can be used to perform an unlogged simulation, typically to
        pre-pace to a (semi-)stable orbit.

        After running this method

        - The simulation time is **not** affected
        - The current state and the default state are updated to the final
          state reached in the simulation.
        - The current state-sensitivities and default state-sensitivities are
          updated to the final state reached in the simulation, except for
          state-sensitivities with respect to initial conditions, which are
          reset (see below).

        Calls to :meth:`reset` after using :meth:`pre` will set the current
        state to this new default state.

        When calculating sensitivities with respect to the initial value of
        some state variable with index ``i``, the sensitivity of that state
        w.r.t. the initial value will be ``1`` at time zero. Similarly, the
        sensitivity of any other state to that initial value at this time will
        be ``0``. Because :meth:`pre()` is used to set a new state for time
        zero, this method will set the current and default state sensitivities
        for initial value sensitivities to 0 or 1 as above, instead of using
        the values reached in the simulation.

        The number of time units to simulate can be set with ``duration``.
        The ``duration`` argument cannot be negative, and special care needs to
        be taken when very small (positive) values are used. For some very
        short (but non-zero) durations, a :class:`myokit.SimulationError` may
        be raised.

        To obtain feedback on the simulation progress, an object implementing
        the :class:`myokit.ProgressReporter` interface can be passed in.
        passed in as ``progress``. An optional description of the current
        simulation to use in the ProgressReporter can be passed in as `msg`.
        """
        duration = float(duration)
        self._run(
            duration, myokit.LOG_NONE, None, None, None, None, None, progress,
            msg)
        self._default_state = list(self._state)
        if self._sensitivities:
            # Reset to time 0, so need to reset initial-value sensitivities
            for i, expr in enumerate(self._sensitivities[1]):
                if isinstance(expr, myokit.InitialValue):
                    self._s_state[i] = [0.0] * len(self._state)
                    self._s_state[i][expr.var().index()] = 1.0
            # Update default state
            self._s_default_state = [list(x) for x in self._s_state]

    def __reduce__(self):
        """
        Pickles this Simulation.

        See: https://docs.python.org/3/library/pickle.html#object.__reduce__
        """
        sens_arg = None
        if self._sensitivities:
            # Don't pass in expressions, as they'll need to pickle the
            # variables as well, which in turn will need a component, model,
            # etc.
            sens_arg = [[x.code() for x in y] for y in self._sensitivities]

        protocols = {
            k: p for k, p in zip(self._pacing_labels, self._protocols)
        }

        return (
            self.__class__,
            (self._model, protocols, sens_arg),
            (
                self._time,
                self._state,
                self._default_state,
                self._s_state,
                self._s_default_state,
                self._tolerance,
                self._dtmin,
                self._dtmax,
            ),
        )

    def reset(self):
        """
        Resets the simulation:

        - The time variable is set to 0
        - The state is set to the default state
        - The state sensitivities are set to the default state sensitivities

        """
        self._time = 0
        self._state = list(self._default_state)
        if self._sensitivities:
            self._s_state = [list(x) for x in self._s_default_state]

    def run(self, duration, log=None, log_interval=None, log_times=None,
            sensitivities=None, apd_variable=None, apd_threshold=None,
            progress=None, msg='Running simulation'):
        """
        Runs a simulation and returns the logged results. Running a simulation
        has the following effects:

        - The internal state is updated to the last state in the simulation.
        - The simulation's time variable is updated to reflect the time
            elapsed during the simulation.

        The number of time units to simulate can be set with ``duration``.
        The ``duration`` argument cannot be negative, and special care needs to
        be taken when very small (positive) values are used. For some very
        short (but non-zero) durations, a :class:`myokit.SimulationError` may
        be raised.

        The method returns a :class:`myokit.DataLog` dictionary that maps
        variable names to lists of logged values. The variables to log can be
        indicated using the ``log`` argument. There are several options for its
        value:

        - ``None`` (default), to log all states.
        - An integer flag or a combination of flags. Options:
          ``myokit.LOG_NONE``, ``myokit.LOG_STATE``, ``myokit.LOG_BOUND``,
          ``myokit.LOG_INTER``, ``myokit.LOG_DERIV`` or ``myokit.LOG_ALL``.
        - A sequence of variable names. To log derivatives, use
          "dot(membrane.V)".
        - A :class:`myokit.DataLog` object. In this case, the new data
          will be appended to the existing log.

        For detailed information about the ``log`` argument, see the function
        :meth:`myokit.prepare_log`.

        By default, every step the solver takes is logged. This is usually
        advantageous, since more points are added exactly at the times the
        system gets more interesting. However, if equidistant points are
        required a ``log_interval`` can be set. Alternatively, the
        ``log_times`` argument can be used to specify logging times directly.

        To get action potential duration (APD) measurements, the simulation can
        be run with threshold crossing detection. To enable this, pass in a
        state variable as ``apd_variable`` and a threshold value as
        ``apd_threshold``.
        *Please note the APD calculation implemented here uses a fixed
        threshold, and so differs from the often used dynamical thresholds such
        as "90% of max(V) - min(V)".*

        To obtain feedback on the simulation progress, an object implementing
        the :class:`myokit.ProgressReporter` interface can be passed in.
        passed in as ``progress``. An optional description of the current
        simulation to use in the ProgressReporter can be passed in as ``msg``.

        Arguments:

        ``duration``
            The time to simulate.
        ``log``
            The variables to log (see detailed description above).
        ``log_interval``
            An optional fixed size log interval. Must be ``None`` if
            ``log_times`` is used. If both are ``None`` every step is logged.
        ``log_times``
            An optional set of pre-determined logging times. Must be ``None``
            if ``log_interval`` is used. If both are ``None`` every step is
            logged.
        ``sensitivities``
            An optional list-of-lists to append the calculated sensitivities
            to.
        ``apd_variable``
            An optional :class:`myokit.Variable` or fully-qualified variable
            name to use in APD calculations. If set, an ``apd_threshold`` must
            also be specified.
        ``apd_threshold``
            An optional (fixed) threshold to use in APD calculations. Must be
            set if and ``apd_variable`` is set, and ``None`` if not.
        ``progress``
            An optional :class:`myokit.ProgressReporter` used to obtain
            feedback about simulation progress.
        ``msg``
            An optional message to pass to any progress reporter.

        By default, this method returns a :class:`myokit.DataLog` containing
        the logged variables.

        However, if sensitivity calculations are enabled a tuple is returned,
        where the first entry is the :class:`myokit.DataLog` and the second
        entry is a matrix of sensitivities ``dy/dx``, where the first index
        specifies the dependent variable ``y``, and the second index specifies
        the independent variable ``x``.

        Finally, if APD calculation is enabled, the method returns a tuple
        ``(log, apds)`` or ``(log, sensitivities, apds)`` where ``apds`` is a
        :class:`myokit.DataLog` with entries ``start`` and ``duration``,
        representing the start and duration of all measured APDs.
        """
        duration = float(duration)
        output = self._run(
            duration, log, log_interval, log_times, sensitivities,
            apd_variable, apd_threshold, progress, msg)
        self._time += duration
        return output

    def _run(self, duration, log, log_interval, log_times, sensitivities,
             apd_variable, apd_threshold, progress, msg):

        # Create benchmarker for profiling and realtime logging
        # Note: When adding profiling messages, write them in past tense so
        # that we can show time elapsed for an operation **that has just
        # completed**.
        if myokit.DEBUG_SP or self._model.binding('realtime') is not None:
            b = myokit.tools.Benchmarker()
            if myokit.DEBUG_SP:
                b.print('PP Entered _run method.')
        else:
            b = None

        # Reset error state
        self._error_state = None

        # Simulation times
        if duration < 0:
            raise ValueError('Simulation time can\'t be negative.')
        tmin = self._time
        tmax = tmin + duration

        # Logging interval (None or 0 = disabled)
        log_interval = 0 if log_interval is None else float(log_interval)
        if log_interval < 0:
            log_interval = 0
        if log_times is not None and log_interval > 0:
            raise ValueError(
                'The arguments `log_times` and `log_interval` cannot be used'
                ' simultaneously.')

        # Check user-specified logging times.
        # (An empty list of log points counts as disabled)
        # Note: Checking of values inside the list (converts to float, is non
        # decreasing) happens in the C code.
        if log_times is not None:
            if len(log_times) == 0:
                log_times = None

        # List of sensitivity matrices
        if self._sensitivities:
            if sensitivities is None:
                sensitivities = []
            else:
                # Must be list (of lists of lists)
                if not isinstance(sensitivities, list):
                    raise ValueError(
                        'The argument `sensitivities` must be None or a list.')
        else:
            sensitivities = None

        # APD measuring
        root_list = None
        root_index = 0
        root_threshold = 0
        if apd_variable is None:
            if apd_threshold is not None:
                raise ValueError(
                    'APD Threshold given but no `apd_variable` specified.')
        else:
            # Get apd variable from this model
            if isinstance(apd_variable, myokit.Variable):
                apd_variable = apd_variable.qname()
            apd_variable = self._model.get(apd_variable)
            if not apd_variable.is_state():
                raise ValueError(
                    'The `apd_variable` must be a state variable.')

            # Set up root finding
            root_list = []
            root_index = apd_variable.index()
            root_threshold = float(apd_threshold)

        # Get progress indication function (if any)
        if progress is None:
            progress = myokit._Simulation_progress
        if progress:
            if not isinstance(progress, myokit.ProgressReporter):
                raise ValueError(
                    'The argument `progress` must be either a'
                    ' subclass of myokit.ProgressReporter or None.')
        if myokit.DEBUG_SP:
            b.print('PP Checked arguments.')

        # Parse log argument
        log = myokit.prepare_log(log, self._model, if_empty=myokit.LOG_ALL)
        if myokit.DEBUG_SP:
            b.print('PP Called prepare_log.')

        # Run simulation
        # The simulation is run only if (tmin + duration > tmin). This is a
        # stronger check than (duration == 0), which will return true even for
        # very short durations (and will cause zero iterations of the
        # "while (t < tmax)" loop below).
        if tmin + duration > tmin:

            # Initial state and sensitivities
            state = list(self._state)
            s_state = None
            if self._sensitivities:
                s_state = [list(x) for x in self._s_state]

            # List to store final bound variables in (for debugging)
            bound = [0, 0, 0] + [0] * len(self._pacing_labels)

            # Initialize
            if myokit.DEBUG_SP:
                b.print('PP Ready to call sim_init.')
            self._sim.sim_init(
                # 0. Initial time
                tmin,
                # 1. Final time
                tmax,
                # 2. Initial and final state
                state,
                # 3. Initial and final state sensitivities
                s_state,
                # 4. Space to store the bound variable values
                bound,
                # 5. Literal values
                list(self._literals.values()),
                # 6. Parameter values
                list(self._parameters.values()),
                # 7. Pacing protocols
                self._protocols,
                # 8. A DataLog
                log,
                # 9. The log interval, or 0
                log_interval,
                # 10. A list of predetermind logging times, or None
                log_times,
                # 11. A list to store calculated sensitivities in
                sensitivities,
                # 12. The state variable index for root finding (only used if
                #     root_list is a list)
                root_index,
                # 13. The threshold for root crossing (can be 0 too, only used
                #     if root_list is a list).
                root_threshold,
                # 14. A list to store calculated root crossing times and
                #     directions in, or None
                root_list,
                # 15. A myokit.tools.Benchmarker or None (if not used)
                b,
                # 16. Boolean/int: 1 if we are logging realtime
                int(self._model.binding('realtime') is not None),
            )
            t = tmin

            # Run
            try:
                if progress:
                    # Loop with feedback
                    with progress.job(msg):
                        r = 1.0 / duration if duration != 0 else 1
                        while t < tmax:
                            t = self._sim.sim_step()
                            if not progress.update(min((t - tmin) * r, 1)):
                                raise myokit.SimulationCancelledError()
                else:
                    # Loop without feedback
                    while t < tmax:
                        t = self._sim.sim_step()

            except ArithmeticError as e:
                # Some CVODE(S) errors are set to raise an ArithmeticError,
                # which users may be able to debug.
                if myokit.DEBUG_SP:
                    b.print('PP Caught ArithmeticError.')

                # Store error state
                self._error_state = state

                # Create long error message
                txt = ['A numerical error occurred during simulation at'
                       ' t = ' + str(t) + '.', 'Last reached state: ']
                txt.extend(['  ' + x for x
                            in self._model.format_state(state).splitlines()])
                txt.append('Inputs for binding:')
                txt.append('  time        = ' + myokit.float.str(bound[0]))
                txt.append('  realtime    = ' + myokit.float.str(bound[1]))
                txt.append('  evaluations = ' + myokit.float.str(bound[2]))
                for i, label in enumerate(self._pacing_labels):
                    txt.append(
                        '  ' + label + ' = ' + myokit.float.str(bound[3 + i])
                    )
                txt.append(str(e))

                # Check if state derivatives can be evaluated in Python, if
                # not, add the error to the error message.
                try:
                    self._model.evaluate_derivatives(state)
                except Exception as en:
                    txt.append(str(en))

                # Raise numerical simulation error
                raise myokit.SimulationError('\n'.join(txt))

            except Exception as e:
                if myokit.DEBUG_SP:
                    b.print('PP Caught exception.')

                # Store error state
                self._error_state = state

                # Cast known CVODE errors as SimulationError
                if 'Function CVode()' in str(e):
                    raise myokit.SimulationError(str(e))

                # Unknown exception: re-raise
                raise
            finally:
                # Clean even after KeyboardInterrupt or other Exception
                self._sim.sim_clean()

            # Update internal state
            # Both lists were newly created, so this is OK.
            self._state = state
            self._s_state = s_state

        # Simulation complete
        if myokit.DEBUG_SP:
            b.print('PP Simulation complete.')

        # Calculate apds
        if root_list is not None:
            st = []
            dr = []
            if root_list:
                roots = iter(root_list)
                time, direction = next(roots)
                tlast = time if direction > 0 else None
                for time, direction in roots:
                    if direction > 0:
                        tlast = time
                    else:
                        st.append(tlast)
                        dr.append(time - tlast)
            apds = myokit.DataLog()
            apds['start'] = st
            apds['duration'] = dr
            if myokit.DEBUG_SP:
                b.print('PP Root-finding data processed.')

        # Return
        if myokit.DEBUG_SP:
            b.print('PP Call to _run() complete. Returning.')
        if self._sensitivities is not None:
            if root_list is not None:
                return log, sensitivities, apds
            else:
                return log, sensitivities
        elif root_list is not None:
            return log, apds
        return log

    def set_constant(self, var, value):
        """
        Changes a model constant. Only literal constants (constants not
        dependent on any other variable) can be changed.

        The constant ``var`` can be given as a :class:`Variable` or a string
        containing a variable qname. The ``value`` should be given as a float.
        """
        # Get variable
        value = float(value)
        if isinstance(var, myokit.Variable):
            var = var.qname()
        var = self._model.get(var)

        # Update value in literal or parameter map
        if var in self._literals:
            self._literals[var] = value
        elif var in self._parameters:
            self._parameters[var] = value
        else:
            raise ValueError(
                'The given variable <' + var.qname() + '> is not a literal.')

        # Update value in internal model: This is required for error handling,
        # when self._model.evaluate_derivatives is called.
        # It also ensures the modified value is retained when pickling.
        self._model.set_value(var, value)

    def set_default_state(self, state):
        """
        Change the default state to ``state``.
        """
        self._default_state = self._model.map_to_state(state)

    def set_max_step_size(self, dtmax=None):
        """
        Sets a maximum step size. To let the solver pick any step size it likes
        use ``dtmax = None``.
        """
        dtmax = 0 if dtmax is None else float(dtmax)
        if dtmax < 0:
            dtmax = 0

        # Store internally
        self._dtmax = dtmax

        # Set in simulation
        self._sim.set_max_step_size(dtmax)

    def set_min_step_size(self, dtmin=None):
        """
        Sets a minimum step size. To let the solver pick any step size it likes
        use ``dtmin = None``.
        """
        dtmin = 0 if dtmin is None else float(dtmin)
        if dtmin < 0:
            dtmin = 0

        # Store internally
        self._dtmin = dtmin

        # Set in simulation
        self._sim.set_min_step_size(dtmin)

    def set_fixed_form_protocol(self, times=None, values=None):
        """
        Sets a :class:`TimeSeriesProtocol` specified by ``times`` and
        ``values`` for the label ``pace``.

        This method is provided for backwards compatibility with older
        versions, please use :meth:`set_protocol` and the
        :class:`TimeSeriesProtocol` class instead.
        """
        # Deprecated on 2023-06-02
        import warnings
        warnings.warn(
            'The method `myokit.Simulation.set_fixed_form_protocol` is '
            'deprecated. It will be removed in future versions of Myokit.'
        )

        if times is None and values is None:
            self.set_protocol(None)
            return
        if times is None:
            raise ValueError('No times given.')
        if values is None:
            raise ValueError('No values given.')

        self.set_protocol(myokit.TimeSeriesProtocol(times, values))

    def set_protocol(self, protocol, label='pace'):
        """
        Set an event-based pacing :class:`Protocol` or a
        :class:`TimeSeriesProtocol` for the given ``label``.

        To remove a previously set binding call this method with ``protocol =
        None``. In this case, the value of any variables bound to ``label``
        will be set to 0.

        The label must be one of the pacing labels set in the constructor.
        """
        try:
            index = self._pacing_labels.index(label)
        except ValueError:
            raise ValueError('Unknown pacing label: ' + str(label))

        # Set new protocol
        self._protocols[index] = None if protocol is None else protocol.clone()

    def __setstate__(self, state):
        """
        Called after unpickling, to set any variables not set by the
        constructor.

        See: https://docs.python.org/3/library/pickle.html#object.__setstate__
        """
        self._time = state[0]
        self._state = state[1]
        self._default_state = state[2]
        self._s_state = state[3]
        self._s_default_state = state[4]

        # The following properties need to be set on the internal simulation
        # object
        self.set_tolerance(*state[5])
        self.set_min_step_size(state[6])
        self.set_max_step_size(state[7])

    def set_state(self, state):
        """
        Sets the current state.
        """
        self._state = self._model.map_to_state(state)

    def set_time(self, time=0):
        """
        Sets the current simulation time.
        """
        self._time = float(time)

    def set_tolerance(self, abs_tol=1e-6, rel_tol=1e-4):
        """
        Sets the solver tolerances. Absolute tolerance is set using
        ``abs_tol``, relative tolerance using ``rel_tol``. For more information
        on these values, see the Sundials CVODES documentation.
        """
        abs_tol = float(abs_tol)
        if abs_tol <= 0:
            raise ValueError('Absolute tolerance must be positive float.')
        rel_tol = float(rel_tol)
        if rel_tol <= 0:
            raise ValueError('Relative tolerance must be positive float.')

        # Store tolerance in Python (for pickling)
        self._tolerance = (abs_tol, rel_tol)

        # Set tolerance in simulation
        self._sim.set_tolerance(abs_tol, rel_tol)

    def state(self):
        """
        Returns the current state.
        """
        return list(self._state)

    def time(self):
        """
        Returns the current simulation time.
        """
        return self._time



#
# Imports a model definition from a CellML file
# Only partial CellML support is provided.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
from lxml import etree

import myokit
import myokit.formats

from myokit.formats.xml import split


class CellMLImporterError(myokit.ImportError):
    """
    Raised if an error occurs when importing CellML.
    """


class CellMLImporter(myokit.formats.Importer):
    """
    This :class:`Importer <myokit.formats.Importer>` imports a model definition
    from CellML.
    """
    def __init__(self, verbose=False):
        super().__init__()

    def model(self, path):
        """
        Reads a CellML file and returns a:class:`myokit.Model`.
        """
        import myokit.formats.cellml as cellml
        import myokit.formats.cellml.v1 as v1
        import myokit.formats.cellml.v2 as v2
        parsers = {
            cellml.NS_CELLML_1_0: v1.CellMLParser,
            cellml.NS_CELLML_1_1: v1.CellMLParser,
            cellml.NS_CELLML_2_0: v2.CellMLParser,
        }

        # Open XML file
        try:
            parser = etree.XMLParser(remove_comments=True)
            tree = etree.parse(path, parser=parser)
        except Exception as e:
            raise CellMLImporterError('Unable to parse XML: ' + str(e))

        # Get root node
        root = tree.getroot()

        # Detect namespace, create appropriate parser
        ns, el = split(root.tag)
        try:
            parser = parsers[ns]
        except KeyError:
            raise CellMLImporterError(
                'Unknown CellML version or not a CellML document at '
                + str(path) + '.')
        parser = parser()

        try:
            # Parse and validate CellML model
            cellml_model = parser.parse(root)
            cellml_model.validate()

            # Create and return Myokit model
            return cellml_model.myokit_model()

        except v1.CellMLParsingError as e:
            raise CellMLImporterError(str(e))
        except v2.CellMLParsingError as e:
            raise CellMLImporterError(str(e))

    def supports_model(self):
        """
        Returns True.
        """
        return True




#
# EasyML expression writer
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import warnings

import myokit

from myokit.formats.python import PythonExpressionWriter


class EasyMLExpressionWriter(PythonExpressionWriter):
    """
    This :class:`ExpressionWriter <myokit.formats.ExpressionWriter>` writes
    equations for variables in EasyML syntax.
    """
    def __init__(self):
        super().__init__()
        self._function_prefix = ''

    #def _ex_name(self, e):
    #def _ex_derivative(self, e):

    def _ex_number(self, e):
        return myokit.float.str(e)

    #def _ex_prefix_plus(self, e):
    #def _ex_prefix_minus(self, e):
    #def _ex_plus(self, e):

    def _ex_minus(self, e):
        if isinstance(e[0], myokit.Exp) and isinstance(e[1], myokit.Number):
            if e[1].eval() == 1:
                return 'expm1(' + self.ex(e[0][0]) + ')'
        if isinstance(e[1], myokit.Exp) and isinstance(e[0], myokit.Number):
            if e[0].eval() == 1:
                return '-expm1(' + self.ex(e[1][0]) + ')'
        return super()._ex_minus(e)

    #def _ex_multiply(self, e):
    #def _ex_divide(self, e):

    def _ex_quotient(self, e):
        return self.ex(myokit.Floor(myokit.Divide(e[0], e[1])))

    def _ex_remainder(self, e):
        return self.ex(myokit.Minus(
            e[0], myokit.Multiply(e[1], myokit.Quotient(e[0], e[1]))))

    def _ex_power(self, e):
        return 'pow(' + self.ex(e[0]) + ', ' + self.ex(e[1]) + ')'

    #def _ex_sqrt(self, e):

    def _ex_sin(self, e):
        warnings.warn('Potentially unsupported function: sin()')
        return super()._ex_sin(e)

    #def _ex_cos(self, e):

    def _ex_tan(self, e):
        warnings.warn('Potentially unsupported function: tan()')
        return super()._ex_tan(e)

    def _ex_asin(self, e):
        warnings.warn('Potentially unsupported function: asin()')
        return super()._ex_asin(e)

    #def _ex_acos(self, e):

    def _ex_atan(self, e):
        warnings.warn('Potentially unsupported function: atan()')
        return super()._ex_atan(e)

    #def _ex_exp(self, e):

    def _ex_log(self, e):
        if len(e) == 1:
            return self._ex_function(e, 'log')
        return '(log(' + self.ex(e[0]) + ') / log(' + self.ex(e[1]) + '))'

    #def _ex_log10(self, e):

    def _ex_floor(self, e):
        warnings.warn('Potentially unsupported function: floor()')
        return super()._ex_floor(e)

    def _ex_ceil(self, e):
        warnings.warn('Potentially unsupported function: ceil()')
        return super()._ex_ceil(e)

    def _ex_abs(self, e):
        return self._ex_function(e, 'fabs')

    def _ex_not(self, e):
        return '!(' + self.ex(e[0]) + ')'

    #def _ex_equal(self, e):
    #def _ex_not_equal(self, e):
    #def _ex_more(self, e):
    #def _ex_less(self, e):
    #def _ex_more_equal(self, e):
    #def _ex_less_equal(self, e):

    def _ex_and(self, e):
        return self._ex_infix_condition(e, 'and')

    def _ex_or(self, e):
        return self._ex_infix_condition(e, 'or')

    def _ex_if(self, e):
        ite = (self.ex(e._i), self.ex(e._t), self.ex(e._e))
        return '(%s ? %s : %s)' % ite

    def _ex_piecewise(self, e):
        s = []
        n = len(e._i)
        for i in range(0, n):
            s.append('(%s ? %s : ' % (self.ex(e._i[i]), self.ex(e._e[i])))
        s.append(self.ex(e._e[n]))
        s.append(')' * n)
        return ''.join(s)




#
# MathML expression writer
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
from lxml import etree

import myokit.formats


class MathMLExpressionWriter(myokit.formats.ExpressionWriter):
    """
    This :class:`ExpressionWriter <myokit.formats.ExpressionWriter>` translates
    Myokit :class:`expressions <myokit.Expression>` to Content MathML or
    Presentation MathML.
    """
    def __init__(self):
        super().__init__()

        # Default mode
        self._pres = False

        # Default lhs conversion function
        def flhs(lhs):
            var = lhs.var()
            if isinstance(var, str):
                # This can happen with time variable of derivative if the
                # proper variable isn't set!
                return var
            return var.qname()
        self._flhs = flhs

        # Default number-to-string conversion function (outputs e.g. "1.0")
        self._fnum = lambda x: myokit.float.str(x.eval()).strip()

        # Default time variable
        self._tvar = myokit.Name('time')

        # Namespaces for element creation
        from myokit.formats.mathml import NS_MATHML_2
        self._nsmap = {None: NS_MATHML_2}

    def set_lhs_function(self, f):
        """
        Sets a naming function, will be called to get the variable name from a
         :class:`myokit.Name` (derivatives will be handled separately).

        The argument ``f`` should be a function that takes a
        :class:`myokit.Name` as input and returns a string.
        """
        self._flhs = f

    def set_mode(self, presentation=False):
        """
        Enables or disables Presentation MathML mode.
        """
        self._pres = bool(presentation)

    def set_time_variable(self, time):
        """
        Sets the time variable to use for this expression writer's derivatives.
        """
        self._tvar = myokit.Name(time)

    def eq(self, eq, element=None):
        """
        Converts an equation to a string.

        The optional argument ``element`` can be used to pass in an
        ``lxml.etree.ElementTree`` element. If given, this element will be
        updated with the generated xml and nothing will be returned.
        """
        if element is None:
            tag = etree.Element('math', nsmap=self._nsmap)
        else:
            tag = element
        if self._pres:
            t = etree.SubElement(tag, 'mrow')
            self.ex(eq.lhs, t)
            x = etree.SubElement(t, 'mo')
            x.text = '='
            self.ex(eq.rhs, t)
        else:
            t = etree.SubElement(tag, 'apply')
            etree.SubElement(t, 'eq')
            self.ex(eq.lhs, t)
            self.ex(eq.rhs, t)
        if element is None:
            # By default et.tostring() will use ascii enconding. This is good,
            # because it creates XML entities (numeric character reference) for
            # special characters.
            enc = 'us-ascii'
            encoded = etree.tostring(tag, encoding=enc)
            return encoded.decode(enc)

    def ex(self, e, element=None):
        """
        Converts an expression to a string.

        The optional argument ``element`` can be used to pass in an
        ``lxml.etree.ElementTree`` element. If given, this element will be
        updated with the generated xml and nothing will be returned.
        """
        if element is None:
            tag = etree.Element('math', nsmap=self._nsmap)
        else:
            tag = element
        self._ex(e, tag)
        if element is None:
            # By default et.tostring() will use ascii enconding. This is good,
            # because it creates XML entities (numeric character reference) for
            # special characters.
            enc = 'us-ascii'
            encoded = etree.tostring(tag, encoding=enc)
            return encoded.decode(enc)

    def _ex(self, e, t):
        """ Writes expression ``e`` to element ``t`` """
        try:
            action = self._op_map[type(e)]
        except KeyError:
            raise ValueError('Unknown expression type: ' + str(type(e)))
        action(e, t)

    def _ex_prefix(self, e, t, cml):
        """
        Exports e as a prefix expression with ContentML representation cml.
        """
        bra = e.bracket(e[0])
        if self._pres:
            row = etree.SubElement(t, 'mrow')
            if bra:
                x = etree.SubElement(row, 'mo')
                x.text = '('
            x = etree.SubElement(row, 'mo')
            x.text = e.operator_rep()
            self._ex(e[0], row)
            if bra:
                x = etree.SubElement(row, 'mo')
                x.text = ')'
        else:
            tag = etree.SubElement(t, 'apply')
            etree.SubElement(tag, cml)
            self._ex(e[0], tag)

    def _ex_infix(self, e, t, cml):
        """
        Exports e as an infix expression with ContentML representation cml.
        """
        if self._pres:
            r = etree.SubElement(t, 'mrow')
            k = etree.SubElement(r, 'mfenced') if e.bracket(e[0]) else r
            self._ex(e[0], k)
            x = etree.SubElement(r, 'mo')
            if isinstance(e, myokit.MoreEqual):
                x.text = '\u2265'
            elif isinstance(e, myokit.LessEqual):
                x.text = '\u2264'
            else:
                x.text = e.operator_rep()
            k = etree.SubElement(r, 'mfenced') if e.bracket(e[1]) else r
            self._ex(e[1], k)
        else:
            a = etree.SubElement(t, 'apply')
            etree.SubElement(a, cml)
            self._ex(e[0], a)
            self._ex(e[1], a)

    def _ex_function(self, e, t, name):
        """ Exports ``e`` as a function called ``name``. """
        if self._pres:
            r = etree.SubElement(t, 'mrow')
            x = etree.SubElement(r, 'mi')
            x.text = name
            r = etree.SubElement(r, 'mfenced')
            for op in e:
                self._ex(op, r)
        else:
            a = etree.SubElement(t, 'apply')
            etree.SubElement(a, name)
            for op in e:
                self._ex(op, a)

    def _ex_name(self, e, t):
        x = etree.SubElement(t, 'mi' if self._pres else 'ci')
        x.text = self._flhs(e)

    def _ex_derivative(self, e, t):
        if self._pres:
            f = etree.SubElement(t, 'mfrac')
            x = etree.SubElement(f, 'mi')
            x.text = 'd' + self._flhs(e[0])
            x = etree.SubElement(f, 'mi')
            x.text = 'dt'
        else:
            a = etree.SubElement(t, 'apply')
            etree.SubElement(a, 'diff')
            self._ex(self._tvar, etree.SubElement(a, 'bvar'))
            self._ex(e[0], a)

    def _ex_number(self, e, t):
        x = etree.SubElement(t, 'mn' if self._pres else 'cn')
        x.text = self._fnum(e)

        if not self._pres:
            try:
                x.text, exp = x.text.split('e', 1)
            except ValueError:
                # Return for overriding
                return x

            exp = int(exp)
            if exp != 0:
                x.attrib['type'] = 'e-notation'
                s = etree.SubElement(x, 'sep')
                s.tail = str(int(exp))

        # Return for overriding
        return x

    def _ex_prefix_plus(self, e, t):
        # Return for overriding
        return self._ex_prefix(e, t, 'plus')

    def _ex_prefix_minus(self, e, t):
        # Return for overriding
        return self._ex_prefix(e, t, 'minus')

    def _ex_plus(self, e, t):
        self._ex_infix(e, t, 'plus')

    def _ex_minus(self, e, t):
        self._ex_infix(e, t, 'minus')

    def _ex_multiply(self, e, t):
        self._ex_infix(e, t, 'times')

    def _ex_divide(self, e, t):
        if self._pres:
            r = etree.SubElement(t, 'mfrac')
            k = etree.SubElement(r, 'mfenced') if e.bracket(e[0]) else r
            self._ex(e[0], k)
            k = etree.SubElement(r, 'mfenced') if e.bracket(e[1]) else r
            self._ex(e[1], k)
        else:
            a = etree.SubElement(t, 'apply')
            etree.SubElement(a, 'divide')
            self._ex(e[0], a)
            self._ex(e[1], a)

    def _ex_quotient(self, e, t):
        self._ex_infix(e, t, 'quotient')

    def _ex_remainder(self, e, t):
        self._ex_infix(e, t, 'rem')

    def _ex_power(self, e, t):
        if self._pres:
            x = etree.SubElement(t, 'msup')
            self._ex(e[0], x)
            self._ex(e[1], x)
        else:
            self._ex_function(e, t, 'power')

    def _ex_sqrt(self, e, t):
        self._ex_function(e, t, 'root')

    def _ex_sin(self, e, t):
        self._ex_function(e, t, 'sin')

    def _ex_cos(self, e, t):
        self._ex_function(e, t, 'cos')

    def _ex_tan(self, e, t):
        self._ex_function(e, t, 'tan')

    def _ex_asin(self, e, t):
        self._ex_function(e, t, 'arcsin')

    def _ex_acos(self, e, t):
        self._ex_function(e, t, 'arccos')

    def _ex_atan(self, e, t):
        self._ex_function(e, t, 'arctan')

    def _ex_exp(self, e, t):
        if self._pres:
            r = etree.SubElement(t, 'msup')
            x = etree.SubElement(r, 'mi')
            x.text = 'e'
            self._ex(e[0], r)
        else:
            a = etree.SubElement(t, 'apply')
            etree.SubElement(a, 'exp')
            self._ex(e[0], a)

    def _ex_log(self, e, t):
        # myokit.log(a)   > ln(a)
        # myokit.log(a,b) > log(b, a)
        # myokit.log10(a) > log(a)
        if self._pres:
            if len(e) == 1:
                r = etree.SubElement(t, 'mrow')
                x = etree.SubElement(r, 'mi')
                x.text = 'ln'
                x = etree.SubElement(r, 'mfenced')
                self._ex(e[0], x)
            else:
                r = etree.SubElement(t, 'mrow')
                s = etree.SubElement(r, 'msub')
                x = etree.SubElement(s, 'mi')
                x.text = 'log'
                self._ex(e[1], s)
                s = etree.SubElement(r, 'mfenced')
                self._ex(e[0], s)
        else:
            if len(e) == 1:
                a = etree.SubElement(t, 'apply')
                etree.SubElement(a, 'ln')
                self._ex(e[0], a)
            else:
                a = etree.SubElement(t, 'apply')
                etree.SubElement(a, 'log')
                x = etree.SubElement(a, 'logbase')
                self._ex(e[1], x)
                self._ex(e[0], a)

    def _ex_log10(self, e, t):
        self._ex_function(e, t, 'log')

    def _ex_floor(self, e, t):
        self._ex_function(e, t, 'floor')

    def _ex_ceil(self, e, t):
        self._ex_function(e, t, 'ceiling')

    def _ex_abs(self, e, t):
        self._ex_function(e, t, 'abs')

    def _ex_not(self, e, t):
        # https://www.w3.org/TR/MathML2/chapter4.html#id.4.4.3.15
        self._ex_prefix(e, t, 'not')

    def _ex_equal(self, e, t):
        self._ex_infix(e, t, 'eq')

    def _ex_not_equal(self, e, t):
        self._ex_infix(e, t, 'neq')

    def _ex_more(self, e, t):
        self._ex_infix(e, t, 'gt')

    def _ex_less(self, e, t):
        self._ex_infix(e, t, 'lt')

    def _ex_more_equal(self, e, t):
        self._ex_infix(e, t, 'geq')

    def _ex_less_equal(self, e, t):
        self._ex_infix(e, t, 'leq')

    def _ex_and(self, e, t):
        self._ex_infix(e, t, 'and')

    def _ex_or(self, e, t):
        self._ex_infix(e, t, 'or')

    def _ex_if(self, e, t):
        self._ex_piecewise(e.piecewise(), t)

    def _ex_piecewise(self, e, t):
        if self._pres:
            w = etree.SubElement(t, 'piecewise')
            for k, cond in enumerate(e._i):
                p = etree.SubElement(w, 'piece')
                self._ex(e._e[k], p)
                self._ex(cond, p)
            p = etree.SubElement(w, 'otherwise')
            self._ex(e._e[-1], p)
        else:
            w = etree.SubElement(t, 'piecewise')
            for k, cond in enumerate(e._i):
                p = etree.SubElement(w, 'piece')
                self._ex(e._e[k], p)
                self._ex(cond, p)
            p = etree.SubElement(w, 'otherwise')
            self._ex(e._e[-1], p)




#
# Right-hand-side equation benchmarking tool.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import os
import platform
import timeit

import myokit

# Location of C source file
SOURCE_FILE = 'rhs.c'


class RhsBenchmarker(myokit.CModule):
    """
    Benchmarks a model's right hand side (rhs).

    To enable partial benchmarking the argument ``variables`` can be used to
    pass in a list of variables whose combined evaluation time can be obtained
    by using :meth:`bench_part()`. Only non-constant, non-bound variables can
    be selected.

    By default, the given list of variables are the only variables included in
    a partial simulation. This behavior can be inverted by setting
    ``exclude_selected=True``. With this setting, all variables except those in
    the given list will be tested.

    A valid myokit model should be provided as the ``model`` argument.
    """
    _index = 0  # Unique id for the generated module

    def __init__(self, model, variables=None, exclude_selected=False):
        super().__init__()

        # Require a valid model
        model.validate()

        # Clone model
        self._model = model.clone()

        # Check given variables
        self._check_variables(variables)

        # Extension module id
        RhsBenchmarker._index += 1
        module_name = 'myokit_RhsBenchmarker_' + str(RhsBenchmarker._index)
        module_name += '_' + str(myokit.pid_hash())

        # Distutils arguments
        args = {
            'module_name': module_name,
            'model': self._model,
            'variables': self._variables,
            'exclude_selected': exclude_selected,
        }
        fname = os.path.join(myokit.DIR_CFUNC, SOURCE_FILE)

        # Define libraries
        libs = []
        if platform.system() != 'Windows':  # pragma: no windows cover
            libs.append('m')

        # Create extension
        self._ext = self._compile(module_name, fname, args, libs)

    def bench_full(self, log, repeats=40000, fastest=False):
        """
        Benchmarks the entire RHS for every position found in the given log
        file.

        The argument ``log`` should point to a :class:`myokit.DataLog` or
        similar containing the values of all state variables and any bound
        variables used in the model.

        The benchmark can be performed in two ways:

        ``fastest=False (default)``
            Each point in the given log is calculated ``repeats`` number of
            times. For each point, the average time it took to do a single
            calculation is returned.
        ``fastest=True``
            Each point in the given log is calculated ``repeats`` number of
            times. For each point, the fastest evaluation time is returned.

        """
        self._check_log(log)
        # Get first and last+1 position to check
        start = 0
        stop = len(log[next(self._model.states()).qname()])
        # Check repeats
        repeats = int(repeats)
        # Create benchmarking function
        bench = timeit.default_timer
        # Run
        times = self._ext.bench_full(bench, log, start, stop, repeats, fastest)
        return times

    def bench_part(self, log, repeats=40000, fastest=False):
        """
        This method benchmarks the combined evaluation time of the RHS for all
        variables passed in when this RhsBenchmarker was created. The combined
        RHS is benchmarked ``repeats`` number of times for every point in the
        given log.

        The argument ``log`` should point to a :class:`myokit.DataLog` or
        similar containing the values of all state variables and any bound
        variables used in the model.

        Only the selected variables RHS will be evaluated, not their
        dependencies or nested children variables. Bound and constant variables
        cannot be selected. If state variables are chosen, their derivative is
        evaluated.

        The benchmark can be performed in two ways:

        ``fastest=False (default)``
            Each point in the given log is calculated ``repeats`` number of
            times. For each point, the average time it took to do a single
            calculation is returned.
        ``fastest=True``
            Each point in the given log is calculated ``repeats`` number of
            times. For each point, the fastest evaluation time is returned.

        """
        self._check_log(log)

        # Get first and last+1 position to check
        start = 0
        stop = len(log[next(self._model.states()).qname()])

        # Check repeats
        repeats = int(repeats)

        # Create benchmarking function
        bench = timeit.default_timer

        # Run
        times = self._ext.bench_part(bench, log, start, stop, repeats, fastest)
        return times

    def mean(self, times):
        """
        Like meth:`mean_std()` but returns only the final mean.
        """
        return self.mean_std(times)[0]

    def mean_std(self, times):
        """
        Removes outliers and then calculates the mean and standard deviation of
        the given benchmarked times.
        """
        import numpy as np
        times = np.array(times, copy=False)
        # Remove outliers twice
        for i in range(0, 2):
            avg = np.mean(times)
            std = np.std(times)
            s3 = std * 3
            times = times[(times > (avg - s3)) * (times < (avg + s3))]
        return np.mean(times), np.std(times)

    def _check_log(self, log):
        """
        Checks the given log for compatibility with this benchmarker.
        """
        n = None
        for state in self._model.states():
            state = state.qname()
            if state not in log:
                raise ValueError(
                    'State variable <' + str(state) + '> not found in'
                    ' simulation log.')
            m = len(log[state])
            if n is None:
                n = m
            elif n != m:
                raise ValueError(
                    'All lists in the simulation log must have the same'
                    ' length. Found entries with size ' + str(n) + ' and '
                    + str(m) + '.')

    def _check_variables(self, variables):
        """
        Checks the given variable list for compatibility with this benchmarker
        and stores a result in self._variables.
        """
        if variables is None:
            self._variables = []
        else:
            self._variables = set()
            for var in variables:
                if isinstance(var, str):
                    # String? Then get variable from model
                    var = self._model.get(var)
                else:
                    # Variable object? Then replace by equivalent form internal
                    # cloned model.
                    var = self._model.get(var.qname())
                if var.is_constant():
                    raise ValueError('Given variables can\'t be constant.')
                if var.is_bound():
                    raise ValueError('Given variables can\'t be bound.')
                # Get same variable from cloned internal version, add to set
                self._variables.add(self._model.get(var.qname()))
            # Create list from set
            self._variables = list(self._variables)



#
# Myokit's version info
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#

# True if this is a release, False for a development version
__release__ = True

# Version as a tuple (major, minor, revision)
#  - Changes to major are rare
#  - Changes to minor indicate new features, possible slight backwards
#    incompatibility
#  - Changes to revision indicate bugfixes, tiny new features
#  - There is no significance to odd/even numbers
__version_tuple__ = 1, 35, 2

# String version of the version number
__version__ = '.'.join([str(x) for x in __version_tuple__])
if not __release__:  # pragma: no cover
    __version_tuple__ += ('dev', )
    __version__ += '.dev'




#!/usr/bin/env python3
#
# Tests the CellML 1.0/1.1 writer
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import re
import unittest

import myokit
import myokit.formats.cellml.v1 as cellml

from myokit.tests import TemporaryDirectory, WarningCollector


class TestCellMLWriter(unittest.TestCase):
    """ Tests for cellml.Writer. """

    def test_cmeta_ids(self):
        # Tests cmeta ids are written for models, components, and variables

        m1 = cellml.Model('m')
        m1.set_cmeta_id('modell')
        c1 = m1.add_component('c')
        c1.set_cmeta_id('commponnent')
        v1 = c1.add_variable('v', 'volt')
        v1.set_cmeta_id('variiable')
        v1.set_initial_value(3)

        xml = cellml.write_string(m1)
        m2 = cellml.parse_string(xml)

        self.assertEqual(m2.cmeta_id(), 'modell')
        self.assertEqual(m2['c'].cmeta_id(), 'commponnent')
        self.assertEqual(m2['c']['v'].cmeta_id(), 'variiable')

    def test_component(self):
        # Tests if components are written

        m1 = cellml.Model('m')
        c = m1.add_component('c')
        d = m1.add_component('d')

        xml = cellml.write_string(m1)
        m2 = cellml.parse_string(xml)

        self.assertIn('c', m2)
        self.assertIn('d', m2)

    def test_connections(self):
        # Tests if connections are written

        m1 = cellml.Model('m')
        c = m1.add_component('c')
        d = m1.add_component('d')
        p = c.add_variable('p', 'mole', public_interface='out')
        q = d.add_variable('q', 'mole', public_interface='in')
        r = c.add_variable('r', 'ampere', public_interface='in')
        s = d.add_variable('r', 'ampere', public_interface='out')
        p.set_initial_value(1)
        s.set_initial_value(2)
        m1.add_connection(p, q)
        m1.add_connection(r, s)

        xml = cellml.write_string(m1)
        m2 = cellml.parse_string(xml)

        p = m2['c']['p']
        q = m2['d']['q']
        r = m2['c']['r']
        s = m2['d']['r']
        self.assertIsNone(p.source())
        self.assertIs(q.source(), p)
        self.assertIs(r.source(), s)
        self.assertIsNone(s.source())

    def test_evaluating_derivatives(self):
        # Writes and then parses a model and compares the derivatives

        m1 = myokit.load_model('example')
        c1 = cellml.Model.from_myokit_model(m1)
        xml = cellml.write_string(c1)
        c2 = cellml.parse_string(xml)
        m2 = c2.myokit_model()

        self.assertEqual(
            m1.get('membrane.V').eval(), m2.get('membrane.V').eval())
        self.assertEqual(m1.get('ina.m').eval(), m2.get('ina.m').eval())
        self.assertEqual(m1.get('ina.h').eval(), m2.get('ina.h').eval())
        self.assertEqual(m1.get('ina.j').eval(), m2.get('ina.j').eval())
        self.assertEqual(m1.get('ica.d').eval(), m2.get('ica.d').eval())
        self.assertEqual(m1.get('ica.f').eval(), m2.get('ica.f').eval())
        self.assertEqual(m1.get('ik.x').eval(), m2.get('ik.x').eval())
        self.assertEqual(m1.get('ica.Ca_i').eval(), m2.get('ica.Ca_i').eval())

    def test_groups(self):
        # Tests if encapsulation relationships are written out.

        # Create a model with
        #      d
        #   a     e
        #       f   c
        #             b
        #
        m1 = cellml.Model('m')
        a = m1.add_component('a')
        b = m1.add_component('b')
        c = m1.add_component('c')
        d = m1.add_component('d')
        e = m1.add_component('e')
        f = m1.add_component('f')
        a.set_parent(d)
        e.set_parent(d)
        f.set_parent(e)
        c.set_parent(e)
        b.set_parent(c)

        xml = cellml.write_string(m1)
        m2 = cellml.parse_string(xml)
        a, b, c, d, e, f = [m2.component(x) for x in 'abcdef']
        self.assertIs(a.parent(), d)
        self.assertIs(e.parent(), d)
        self.assertIs(f.parent(), e)
        self.assertIs(c.parent(), e)
        self.assertIs(b.parent(), c)

    def test_initial_value_representation(self):
        # Test the way initial values are represented in generated CellML code

        def find(xml):
            i = xml.index(b'initial_value')
            i += 15
            j = xml.index(b'"', i)
            return xml[i:j].decode()

        m = cellml.Model('m', '1.0')
        c = m.add_component('c')
        p = c.add_variable('p', 'mole')

        p.set_initial_value(1.234)
        x = find(cellml.write_string(m))
        self.assertEqual(x, '1.234')

        p.set_initial_value(1e-6)
        x = find(cellml.write_string(m))
        self.assertEqual(x, '1e-06')

        p.set_initial_value(1e9)
        x = find(cellml.write_string(m))
        self.assertEqual(x, '1.00000000000000000e+09')

        # String e+00
        p.set_initial_value(1.23424352342423)
        x = find(cellml.write_string(m))
        self.assertEqual(x, '1.23424352342422994')

    def test_maths(self):
        # Test maths is written

        self._test_maths('1.0')
        self._test_maths('1.1')

    def _test_maths(self, version):
        # Test maths is written in either 1.0 or 1.1

        # Create model
        m1 = cellml.Model('m', version)
        c1 = m1.add_component('c')
        p1 = c1.add_variable('p', 'mole')
        p1.set_initial_value(2)
        q1 = c1.add_variable('q', 'dimensionless')
        r1 = c1.add_variable('r', 'second')
        r1.set_is_state(True)
        r1.set_initial_value(0.1)
        t1 = c1.add_variable('t', 'second')
        m1.set_free_variable(t1)

        # Add component without maths
        d1 = m1.add_component('d')
        s1 = d1.add_variable('s', 'volt')
        s1.set_initial_value(1.23)

        # Add two rhs equations
        eq1 = myokit.Plus(myokit.Number(3, myokit.units.mole), myokit.Name(p1))
        # Note: Numbers without units become dimensionless in CellML
        er1 = myokit.Power(myokit.Name(q1), myokit.Number(2))
        q1.set_rhs(eq1)
        r1.set_rhs(er1)

        # Write and read
        xml = cellml.write_string(m1)
        m2 = cellml.parse_string(xml)

        # Check results
        p2, q2, r2, s2 = m2['c']['p'], m2['c']['q'], m2['c']['r'], m2['d']['s']
        subst = {
            myokit.Name(p1): myokit.Name(p2),
            myokit.Name(q1): myokit.Name(q2),
        }
        eq2 = eq1.clone(subst)
        er2 = er1.clone(subst)
        self.assertEqual(q2.rhs(), eq2)
        self.assertEqual(r2.rhs(), er2)
        self.assertEqual(s2.initial_value(), 1.23)
        self.assertFalse(p2.is_state())
        self.assertFalse(q2.is_state())
        self.assertTrue(r2.is_state())
        self.assertFalse(s2.is_state())
        self.assertIs(m2.free_variable(), m2['c']['t'])

    def test_model(self):
        # Tests model writing

        m1 = cellml.Model('model_name')
        xml = cellml.write_string(m1)
        m2 = cellml.parse_string(xml)

        self.assertEqual(m2.name(), 'model_name')
        self.assertEqual(len(m2), 0)

    def test_ordering(self):
        # Tests ordering of components and variables

        # Check component ordering
        m = cellml.Model('m')
        c = m.add_component('C')
        a = m.add_component('A')
        d = m.add_component('D')
        b = m.add_component('B')

        xml = cellml.write_string(m)
        reg = re.compile(br'<component [^>]*name="[\w]+"')
        items = reg.findall(xml)
        items_sorted = list(sorted(items))
        self.assertEqual(items, items_sorted)

        # Check variable ordering
        m = cellml.Model('m')
        c = m.add_component('C')
        r = c.add_variable('r', 'volt')
        p = c.add_variable('p', 'ampere')
        s = c.add_variable('s', 'kilogram')
        q = c.add_variable('q', 'mole')
        r.set_initial_value(1)
        p.set_initial_value(1)
        s.set_initial_value(1)
        q.set_initial_value(1)

        xml = cellml.write_string(m)
        reg = re.compile(br'<variable [^>]*name="[\w]+"')
        items = reg.findall(xml)
        items_sorted = list(sorted(items))
        self.assertEqual(items, items_sorted)

    def test_oxmeta_annotations(self):
        # Tests if oxmeta annotations are written

        # Create mini model
        m = cellml.Model('m')
        c = m.add_component('c')
        v = c.add_variable('v', 'volt')
        v.set_rhs(myokit.Number(0, myokit.units.volt))

        # Test no RDF is written without oxmeta or cmeta
        xml = cellml.write_string(m)
        self.assertNotIn(b'rdf:RDF', xml)

        v.meta['oxmeta'] = 'membrane_voltage'
        xml = cellml.write_string(m)
        self.assertNotIn(b'rdf:RDF', xml)

        del v.meta['oxmeta']
        v.set_cmeta_id('vvv')
        xml = cellml.write_string(m)
        self.assertNotIn(b'rdf:RDF', xml)

        # Test it is written if both are set
        v.meta['oxmeta'] = 'membrane_voltage'
        xml = cellml.write_string(m)
        self.assertIn(b'rdf:RDF', xml)

        # Test reading it again with the parser
        m2 = cellml.parse_string(xml)
        v2 = m2['c']['v']
        self.assertEqual(v2.cmeta_id(), 'vvv')
        self.assertIn('oxmeta', v2.meta)
        self.assertEqual(v2.meta['oxmeta'], 'membrane_voltage')

    def test_units(self):
        # Test writing of units

        u1 = myokit.parse_unit('kg*m^2/mole^3 (0.123)')
        u2 = myokit.parse_unit('1 (4.56)')

        m1 = cellml.Model('mmm')
        m1.add_units('flibbit', u1)
        m1.add_units('special_volt', myokit.units.Volt)
        c = m1.add_component('ccc')
        c.add_units('flibbit', u2)
        p = c.add_variable('p', 'flibbit')
        p.set_rhs(myokit.Number(1, u2))
        d = m1.add_component('ddd')
        q = d.add_variable('q', 'flibbit')
        q.set_rhs(myokit.Number(2, u1))

        xml = cellml.write_string(m1)
        m2 = cellml.parse_string(xml)

        p, q = m2['ccc']['p'], m2['ddd']['q']
        self.assertEqual(p.units().name(), 'flibbit')
        self.assertEqual(q.units().name(), 'flibbit')
        self.assertEqual(p.units().myokit_unit(), u2)
        self.assertEqual(q.units().myokit_unit(), u1)
        self.assertEqual(
            m2.find_units('special_volt').myokit_unit(), myokit.units.volt)

    def test_variable(self):
        # Tests writing of variables

        m1 = cellml.Model('m')
        c = m1.add_component('c')
        p = c.add_variable('p', 'mole')
        q = c.add_variable('q', 'kelvin', public_interface='in')
        r = c.add_variable('r', 'ampere', private_interface='out')
        p.set_initial_value(1)

        with WarningCollector():
            xml = cellml.write_string(m1)
            m2 = cellml.parse_string(xml)

        p, q, r = m2['c']['p'], m2['c']['q'], m2['c']['r']
        self.assertEqual(p.units().name(), 'mole')
        self.assertEqual(q.units().name(), 'kelvin')
        self.assertEqual(r.units().name(), 'ampere')
        self.assertEqual(p.public_interface(), 'none')
        self.assertEqual(q.public_interface(), 'in')
        self.assertEqual(r.public_interface(), 'none')
        self.assertEqual(p.private_interface(), 'none')
        self.assertEqual(q.private_interface(), 'none')
        self.assertEqual(r.private_interface(), 'out')
        self.assertEqual(p.initial_value(), 1)
        self.assertEqual(q.initial_value(), None)
        self.assertEqual(r.initial_value(), None)

    def test_version(self):
        # Tests if both versions can be written

        # Version 1.0
        m1 = cellml.Model('m', version='1.0')
        xml = cellml.write_string(m1)
        m2 = cellml.parse_string(xml)
        self.assertEqual(m2.version(), '1.0')

        # Version 1.1
        m1 = cellml.Model('m', version='1.1')
        xml = cellml.write_string(m1)
        m2 = cellml.parse_string(xml)
        self.assertEqual(m2.version(), '1.1')

    def test_write_file(self):
        # Tests write_file

        m1 = cellml.Model('ernie')
        with TemporaryDirectory() as d:
            path = d.path('test.cellml')
            cellml.write_file(path, m1)
            m2 = cellml.parse_file(path)

        self.assertEqual(m2.name(), 'ernie')
        self.assertEqual(len(m2), 0)

    def test_write_string(self):
        # Tests write_string

        m1 = cellml.Model('ernie')
        xml = cellml.write_string(m1)
        m2 = cellml.parse_string(xml)

        self.assertEqual(m2.name(), 'ernie')
        self.assertEqual(len(m2), 0)


if __name__ == '__main__':
    import warnings
    warnings.simplefilter('always')
    unittest.main()



#
# CellML 2.0 Writer: Writes a CellML Model to disk
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
from lxml import etree

import myokit
import myokit.formats.cellml as cellml


def write_file(path, model):
    """
    Writes a CellML 2.0 model to the given path.
    """
    return CellMLWriter().write_file(path, model)


def write_string(model):
    """
    Writes a CellML 2.0 model to a string and returns it.
    """
    return CellMLWriter().write_string(model)


class CellMLWriter:
    """
    Writes CellML 2.0 documents.
    """

    # List of si unit names corresponding to myokit.Unit exponents
    _exp_si = None

    # Note: Most items are sorted, to get closer to a 'canonical form' CellML
    # document: https://github.com/cellml/libcellml/issues/289

    def _component(self, parent, component, voi):
        """
        Adds an etree ``Element`` to a ``parent`` element, representing the
        given CellML ``component``.
        """

        # Create component element
        element = etree.SubElement(parent, 'component')
        element.attrib['name'] = component.name()

        # Add variables
        for variable in sorted(component.variables(), key=_name):
            self._variable(element, variable)

        # Add maths
        self._maths(element, component, voi)

    def _connections(self, parent, model):
        """
        Adds ``connections`` elements to the given ``parent`` tag for all
        connections needed in ``model``.
        """

        # Create a dictionary where each key is a pair
        # ``(component_1, component_2)`` ordered alphabetically by name, and
        # each value is a (non-empty) list of ``(variable_1, variable_2)``
        # pairs, such that ``variable_1`` is in ``component_1`` and
        # ``variable_2`` is in ``component_2``.
        connections = {}

        # Scan all connections
        for v1, v2 in model.connections():
            c1 = v1.component()
            c2 = v2.component()

            # Order components alphabetically
            if c2.name() < c1.name():
                c1, c2 = c2, c1
                v1, v2 = v2, v1

            # Store the connection
            try:
                vrs = connections[(c1, c2)]
            except KeyError:
                connections[(c1, c2)] = vrs = set()
            vrs.add((v1, v2))

        # Add elements
        for cs, vs in sorted(connections.items(), key=lambda x: _names(x[0])):

            # Add connection
            connection = etree.SubElement(parent, 'connection')
            connection.attrib['component_1'] = cs[0].name()
            connection.attrib['component_2'] = cs[1].name()

            # Add variables
            for v1, v2 in sorted(vs, key=_names):
                map_variables = etree.SubElement(connection, 'map_variables')
                map_variables.attrib['variable_1'] = v1.name()
                map_variables.attrib['variable_2'] = v2.name()

    def _encapsulation(self, parent, model):
        """
        Adds an ``encapsulation`` element to ``parent`` that contains all
        encapsulation relationships present in the given ``model``.
        """

        # Check if this model has encapsulation relationships, if not: return
        has_encapsulation = False
        for component in model:
            if component.parent() is not None:
                has_encapsulation = True
                break
        if not has_encapsulation:
            return

        # Create element
        element = etree.SubElement(parent, 'encapsulation')

        # Add all unencapsulated components that have children
        for component in sorted(model, key=_name):
            if component.parent() is None and component.has_children():
                self._encapsulation_component_ref(element, component)

    def _encapsulation_component_ref(self, parent, component):
        """
        Adds a ``component_ref`` element for ``component`` to ``parent``, and
        then recurses to any children of the component.
        """

        # Create element
        element = etree.SubElement(parent, 'component_ref')
        element.attrib['component'] = component.name()

        # Add children
        for child in sorted(component.children(), key=_name):
            self._encapsulation_component_ref(element, child)

    def _maths(self, parent, component, voi):
        """
        Adds a ``math`` element to the given ``parent`` containing the maths
        for the variables in ``component``.

        Derivatives will be written with reference to a local variable in the
        connected variable set of the variable of integration ``voi``.
        """

        # Test if this component has maths
        has_maths = False
        for v in component:
            if v.rhs() is not None:
                has_maths = True
                break
        if not has_maths:
            return

        # Find local alias of the variable of integration (in valid models,
        # this will always be set if states are present in this component).
        if voi is not None:
            if voi.component() is not component:
                for var in voi.connected_variables():
                    if var.component() is component:
                        voi = var
                        break

        # Create expression writer for this component
        from myokit.formats.cellml import CellMLExpressionWriter
        model = component.model()
        ewriter = CellMLExpressionWriter(model.version())
        ewriter.set_lhs_function(lambda x: x.var().name())
        ewriter.set_unit_function(lambda x: model.find_units_name(x))
        if voi is not None:
            ewriter.set_time_variable(voi)

        # Reset default namespace to MathML namespace
        nsmap = {None: cellml.NS_MATHML}
        #if component.model().version() == '2.0':
        nsmap['cellml'] = cellml.NS_CELLML_2_0

        # Create math elements
        math = etree.SubElement(parent, 'math', nsmap=nsmap)

        # Add maths for variables
        for variable in sorted(component, key=_name):
            if variable.equation_variable() is variable:
                ewriter.eq(variable.equation(), math)

    def _model(self, model):
        """
        Returns an etree ``Element`` representing the given ``model``.
        """

        # Load correct namespaces
        version = model.version()
        # if version == '2.0':
        namespaces = {None: cellml.NS_CELLML_2_0}
        namespaces['cellml'] = namespaces[None]

        # Create model element
        element = etree.Element('model', nsmap=namespaces)

        # Set model name
        element.attrib['name'] = model.name()

        # Add units
        for units in sorted(model.units(), key=_name):
            self._units(element, units)

        # Add components
        voi = model.variable_of_integration()
        for component in sorted(model, key=_name):
            self._component(element, component, voi)

        # Add connections
        self._connections(element, model)

        # Add encapsulation
        self._encapsulation(element, model)

        # Return model element
        return element

    def _units(self, parent, units):
        """
        Adds a ``units`` element to ``parent``, for the given CellML units
        object.
        """
        # Get unit exponents on first call
        if self._exp_si is None:
            from myokit.formats.cellml import v2
            self._exp_si = [
                v2.Units._si_units_r[x] for x in myokit.Unit.list_exponents()]

        # Create units element
        element = etree.SubElement(parent, 'units')
        element.attrib['name'] = units.name()

        # Get myokit unit
        myokit_unit = units.myokit_unit()

        # Add unit row for each of the 7 SI units needed to make up this unit
        rows = []
        for k, e in enumerate(myokit_unit.exponents()):
            if e != 0:
                row = etree.SubElement(element, 'unit')
                row.attrib['units'] = self._exp_si[k]
                if e != 1:
                    row.attrib['exponent'] = str(e)
                rows.append(row)

        # Handle dimensionless units with a multiplier
        if not rows:
            row = etree.SubElement(element, 'unit')
            row.attrib['units'] = 'dimensionless'
            rows.append(row)

        # Add multiplier or prefix to first row
        multiplier = myokit_unit.multiplier()
        if multiplier != 1:
            if myokit.float.eq(multiplier, int(multiplier)):
                rows[0].attrib['multiplier'] = str(int(multiplier))
            else:
                rows[0].attrib['multiplier'] = myokit.float.str(
                    multiplier).strip()

    def _variable(self, parent, variable):
        """
        Adds a ``variable`` element to ``parent`` for the variable represented
        by :class:`myokit.formats.cellml.v2.Variable` ``variable``.
        """

        # Create element
        element = etree.SubElement(parent, 'variable')
        element.attrib['name'] = variable.name()

        # Add units
        element.attrib['units'] = variable.units().name()

        # Add interfaces
        if variable.interface() != 'none':
            element.attrib['interface'] = variable.interface()

        # Add initial value
        if variable is variable.initial_value_variable():
            value = myokit.float.str(variable.initial_value()).strip()
            if value[-4:] == 'e+00':
                value = value[:-4]
            element.attrib['initial_value'] = value

    def write(self, model):
        """
        Takes a :class:`myokit.formats.cellml.v2.Model` as input, and
        creates an ElementTree that represents it.

        If the model contains any variables that have an `oxmeta` meta data
        property, this will be annotated with RDF tags suitable for use with
        the Cardiac Electrophysiology Web Lab.
        """
        # Validate model
        model.validate()

        try:
            # Variables with an oxmeta annotation and a cmeta:id will get an
            # annotation suitable for use with the web lab. This dict maps
            # cmeta:id strings to oxmeta annotations.
            self._oxmeta_variables = {}

            # Temporarily store variable of integration (in valid models, this
            # is always set if states are used).
            self._time = model.variable_of_integration()

            # Create model element (with children)
            element = self._model(model)

            # Wrap in ElementTree and return
            return etree.ElementTree(element)

        finally:
            # Delete any temporary properties
            del self._oxmeta_variables, self._time

    def write_file(self, path, model):
        """
        Takes a :class:`myokit.formats.cellml.v2.Model` as input, and writes it
        to the given ``path``.

        See :meth:`write()` for details.
        """

        # Create ElementTree
        tree = self.write(model)

        # Write to disk
        tree.write(
            path,
            encoding='utf-8',
            method='xml',
            xml_declaration=True,
            pretty_print=True,
        )
        # Note: Can use method='c14n' to get canonical output, but that also
        # removes the xml declaration, which I quite like having!

    def write_string(self, model):
        """
        Takes a :class:`myokit.formats.cellml.v2.Model` as input, and converts
        it to an XML string.

        See :meth:`write()` for details.
        """

        # Create ElementTree
        tree = self.write(model)

        # Write to string
        return etree.tostring(
            tree,
            encoding='utf-8',
            method='xml',
            pretty_print=True,
        )


def _name(x):
    """ Sort by name. """
    return x.name()


def _names(x):
    """ Sort by names. """
    return x[0].name(), x[1].name()




#!/usr/bin/env python3
#
# Tests the RhsBenchmarker
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import unittest

import numpy as np

import myokit


class RhsBenchmarkerTest(unittest.TestCase):
    """
    Tests :class:`RhsBenchmarker`.
    """
    def test_simple(self):
        # Test basic functionality.

        # Create test model
        m = myokit.Model('test')
        c = m.add_component('c')
        t = c.add_variable('time')
        t.set_rhs('0')
        t.set_binding('time')
        v = c.add_variable('V')
        v.set_rhs('0')
        v.promote(-80.1)
        x = c.add_variable('x')
        x.set_rhs('exp(V)')
        m.validate()

        # Create simulation log
        log = myokit.DataLog()
        log['c.time'] = np.zeros(10)
        log['c.V'] = np.linspace(-80.0, 50.0, 10)

        # Number of repeats
        repeats = 10

        # Run
        x.set_rhs('1 / (7 * exp((V + 12) / 35) + 9 * exp(-(V + 77) / 6))')
        b = myokit.RhsBenchmarker(m, [x])
        t = b.bench_full(log, repeats)
        t = b.bench_part(log, repeats)
        # No errors = pass

        # Get mean and std after outlier removal
        mean = b.mean(t)
        mean, std = b.mean_std(t)

    def test_bad_log(self):
        # Test error handling when unsuitable log is used.

        # Create test model
        m = myokit.Model('test')
        c = m.add_component('c')
        t = c.add_variable('time')
        t.set_rhs('0')
        t.set_binding('time')
        v = c.add_variable('V')
        v.set_rhs('0')
        v.promote(-80.1)
        w = c.add_variable('W')
        w.set_rhs(0)
        w.promote(10)
        x = c.add_variable('x')
        x.set_rhs('exp(V) + W')
        m.validate()

        # Create benchmarker
        b = myokit.RhsBenchmarker(m, [x])

        # Missing state variable
        d = myokit.DataLog()
        d['c.time'] = np.zeros(10)
        #d['c.V'] = np.linspace(-80.0, 50.0, 10)
        d['c.W'] = np.linspace(0.0, 10.0, 10)
        self.assertRaisesRegex(
            ValueError, 'State variable <c.V> not found', b.bench_full, d, 1)

        # Different size log entries
        d = myokit.DataLog()
        d['c.time'] = np.zeros(1000)
        d['c.V'] = np.linspace(-80.0, 50.0, 10)
        d['c.W'] = np.linspace(0.0, 10.0, 11)
        self.assertRaisesRegex(
            ValueError, 'same length', b.bench_full, d, 1)

    def test_creation(self):
        # Test Benchmarker creation.
        # Create test model
        m = myokit.Model('test')
        c = m.add_component('c')
        t = c.add_variable('time')
        t.set_rhs('0')
        t.set_binding('time')
        v = c.add_variable('V')
        v.set_rhs('0')
        v.promote(-80.1)
        x = c.add_variable('x')
        x.set_rhs('exp(V)')
        y = c.add_variable('y')
        y.set_rhs(3)
        m.validate()

        # Create benchmarker without variables
        myokit.RhsBenchmarker(m)

        # Create with objects
        myokit.RhsBenchmarker(m, [x])

        # Create with strings
        myokit.RhsBenchmarker(m, ['c.x'])

        # Cannot create with constants
        self.assertRaisesRegex(
            ValueError, 'constant', myokit.RhsBenchmarker, m, [y])

        # Cannot create with bound variables
        self.assertRaisesRegex(
            ValueError, 'bound', myokit.RhsBenchmarker, m, [t])


if __name__ == '__main__':
    unittest.main()



#
# Easy access to predefined myokit units.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import myokit


def _add(name, unit, quantifiable=False, output=False):
    """
    Adds the given unit to the local namespace and registers it with the myokit
    unit system.
    """
    globals()[name] = unit
    myokit.Unit.register(name, unit, quantifiable, output)


def _rep(name):
    unit = myokit.parse_unit(name)
    myokit.Unit.register_preferred_representation(name, unit)


#
# SI myokit.Units
#
# Dimensionless
_add('dimensionless', myokit.Unit())
# Angles
_add('rad', dimensionless)
_add('radian', dimensionless)
_add('sr', dimensionless)
_add('steradian', dimensionless)
# Basic SI units
_add('kg', myokit.Unit(
    [1, 0, 0, 0, 0, 0, 0], 3), quantifiable=False, output=True)
_add('m', myokit.Unit([0, 1, 0, 0, 0, 0, 0]), quantifiable=True, output=True)
_add('s', myokit.Unit([0, 0, 1, 0, 0, 0, 0]), quantifiable=True, output=True)
_add('A', myokit.Unit([0, 0, 0, 1, 0, 0, 0]), quantifiable=True, output=True)
_add('K', myokit.Unit([0, 0, 0, 0, 1, 0, 0]), quantifiable=True, output=True)
_add('cd', myokit.Unit([0, 0, 0, 0, 0, 1, 0]), quantifiable=True, output=True)
_add('mol', myokit.Unit([0, 0, 0, 0, 0, 0, 1]), quantifiable=True, output=True)
# SI unit aliases
_add('kilogram', kg)
_add('metre', m)
_add('meter', m)
_add('second', s)
_add('ampere', A)
_add('Ampere', A)
_add('kelvin', K)
_add('Kelvin', K)
_add('candela', cd)
_add('mole', mol)
# Derived SI units
# The SI units are added in lowercase, as specified by the SI
# The SI units that are derived from names are also added capitalized, as this
# is a common enough mistake...
_add('Hz', 1 / s, quantifiable=True, output=True)
_add('hertz', Hz)       # frequency
_add('Hertz', Hz)
_add('N', kg * m / s**2, quantifiable=True, output=True)
_add('newton', N)       # force
_add('Newton', N)
_add('Pa', N / m**2, quantifiable=True, output=True)
_add('pascal', Pa)      # pressure
_add('Pascal', Pa)
_add('J', N * m, quantifiable=True, output=True)
_add('joule', J)        # energy
_add('Joule', J)
_add('W', J / s, quantifiable=True, output=True)
_add('watt', W)         # power
_add('Watt', W)
_add('C', s * A, quantifiable=True, output=True)
_add('coulomb', C)      # charge
_add('Coulomb', C)
_add('V', J / C, quantifiable=True, output=True)
_add('volt', V)         # electric potential
_add('Volt', V)
_add('F', C / V, quantifiable=True, output=True)
_add('farad', F)        # electric capacitance
_add('Farad', F)
_add('Ohm', V / A, quantifiable=True, output=True)
_add('ohm', Ohm)        # electric resistance
_add('R', Ohm, quantifiable=True)   # Legacy: Deprecated on May 26th 2020
_add('S', 1 / R, quantifiable=True, output=True)
_add('siemens', S)      # siemens (electric conductivity)
_add('Siemens', S)
_add('Wb', J / A, quantifiable=True, output=True)
_add('weber', Wb)       # magnetic flux
_add('Weber', Wb)
_add('T', Wb / m**2, quantifiable=True, output=True)
_add('tesla', T)        # magnetic flux
_add('Tesla', T)
_add('H', Wb / A, quantifiable=True, output=True)
_add('henry', H)        # magnetic field strength
_add('Henry', H)
_add('lm', cd * sr, quantifiable=True, output=True)
_add('lumen', lm)       # Luminous flux
_add('lx', lm / m**2, quantifiable=True, output=True)
_add('lux', lx)         # Illuminance
_add('Bq', 1 / s, quantifiable=True, output=True)
_add('becquerel', Bq)    # radioactive decays / s
_add('Becquerel', Bq)
_add('Gy', J / kg, quantifiable=True, output=True)
_add('gray', Gy)        # absorbed dose of ionizing radiation
_add('Gray', Gy)
_add('Sv', J / kg, quantifiable=True, output=True)
_add('sievert', Sv)     # equivalent dose of ionizing radiation
_add('Sievert', Sv)
_add('kat', mol / s, quantifiable=True, output=True)
_add('katal', kat)      # catalytic activity
# Commonly quantified SI units
_add('g', kg * 1e-3, quantifiable=True, output=True)
_add('km', m * 1e3, output=True)
_add('hm', m * 1e2, output=False)
_add('dm', m * 1e-1, output=False)
_add('cm', m * 1e-2, output=True)
_add('mm', m * 1e-3, output=True)
_add('um', m * 1e-6, output=True)
_add('nm', m * 1e-9, output=True)
_add('mmol', mol * 1e-3, output=True)
_add('umol', mol * 1e-6, output=True)
_add('nmol', mol * 1e-9, output=True)
_add('ms', s * 1e-3, output=True)
_add('us', s * 1e-6, output=True)
_add('mV', V * 1e-3, output=True)
_add('uV', V * 1e-6, output=True)
_add('mA', A * 1e-3, output=True)
_add('uA', A * 1e-6, output=True)
_add('nA', A * 1e-9, output=True)
_add('pA', A * 1e-12, output=True)
_add('mF', F * 1e-3, output=True)
_add('uF', F * 1e-6, output=True)
_add('nF', F * 1e-9, output=True)
_add('pF', F * 1e-12, output=True)
_add('mS', S * 1e-3, output=True)
_add('uS', S * 1e-6, output=True)
_add('nS', S * 1e-9, output=True)
_add('pS', S * 1e-12, output=True)
_add('MOhm', Ohm * 1e6, output=True)
_add('GOhm', Ohm * 1e9, output=True)

#
# Common non-SI units
#
# Common time measures
_add('minute', s * 60)
_add('hour', minute * 60)
_add('hr', hour)
_add('h', hour)
_add('day', hour * 24)
_add('week', day * 7)
_add('wk', week)
_add('julian_year', day * 365.25)
_add('year', julian_year)
_add('yr', year)
# Distance
_add('parsec', m * 30.857e15)
_add('pc', parsec)
_add('light_year', m * 9.4605284e15)    # based on julian year
_add('ly', light_year)
_add('angstrom', m / 1e10)
# Area
_add('hectare', hm**2)      # ha
# Volume
_add('L', dm**3, quantifiable=True, output=True)
_add('liter', L)
_add('litre', L)
_add('cc', cm**3)
# Weight / Mass
_add('Da', kg * 1.66053886e-27, quantifiable=True, output=True)
_add('dalton', Da)
_add('tonne', kg * 1e3)
# Pressure
_add('bar', Pa * 1e5)
_add('atm', 1.01325 * bar)
_add('torr', atm / 760)     # Approx. equal to 1 mmHg
# Concentration
_add('M', mol / L, quantifiable=True, output=True)
_add('molar', M)
#
# Imperial units (international version)
#
# For imperial units its customary to use the plural (7 feet, 12 yards etc)
#
# 1959: Yard defined as exactly 0.9144m. Others can be related to yard. But
# 0.9144 / 3 / 12', 0.0254 exactly, so you can build it from a 2.54cm inch as
# well.
#
_add('inches', cm * 2.54)       # in, or "
_add('thou', inches / 1000)
_add('feet', inches * 12)       # ft, or '
_add('yards', feet * 3)         # yd
_add('chains', yards * 22)
_add('furlongs', chains * 10)
_add('miles', furlongs * 8)     # 8 * 10 * 22', 1760 yards
_add('leagues', miles * 3)
_add('nautical_miles', feet * 6080)
_add('cables', nautical_miles / 10)
_add('fathoms', cables / 100)   # commonly rounded to 6 foot
# Surveying units:
_add('rods', chains / 4)
_add('links', rods / 25)
_add('perches', rods * rods)
_add('roods', furlongs * rods)
_add('acres', furlongs * chains)
# Volumes
_add('gallons', 4.54609 * litre)
_add('quarts', gallons / 4)
_add('pints', quarts / 2)
_add('half_pints', pints / 2)
_add('gills', pints / 4)
_add('fluid_ounces', pints / 20)
_add('floz', fluid_ounces)
# Mass (but not for precious metals)
_add('pounds', 453.59237 * g)   # By definition
_add('lb', pounds)
_add('ounces', pounds / 16)
_add('oz', ounces)
_add('drachms', ounces / 16)
_add('grains', pounds / 7000)   # Originally the weight of a grain of cereal
_add('stones', pounds * 14)
_add('quarters', stones * 2)            # 1/4 of a hundredweight
_add('hundredweights', quarters * 4)    # 8 stone, 112 pounds
_add('imp_tons', 20 * hundredweights)
#
# Not-so-common units
#
_add('dog_year', julian_year / 7)
_add('forthnight', 14 * day)
_add('smoot', (5 + 7 / 12) * feet)
#
# Preferred representation for (un)common electrophysiological units
#
# Sizes and speeds
_rep('1/ms')
_rep('1/m')
_rep('1/cm')
_rep('1/mm')
_rep('cm^2')
_rep('cm^3')
_rep('cm^2/s')
_rep('cm^3/s')
_rep('cm/s')
_rep('cm/ms')
_rep('mm^2')
_rep('mm^3')
_rep('um^2')
_rep('um^3')
_rep('nm^2')
_rep('nm^3')
# _rep('m/s')
# Litres
_rep('mL')
_rep('uL')
_rep('nL')
_rep('pL')
# Concentrations
_rep('M')
_rep('mM')
_rep('mM^2')
_rep('mM^3')
_rep('mM^4')
_rep('uM')
_rep('1/mM/ms')
_rep('1/mM^2/ms')
_rep('mM/ms')
_rep('mM/mV/ms')
# Voltages
_rep('1/mV')
_rep('1/mV^2')
_rep('1/mV/ms')
_rep('ms/mV')
_rep('mV^2')
# Currents
_rep('A/F')
_rep('nA/cm^2')
_rep('uA/cm^2')
_rep('pA/cm^2')
# Conductances and permeability
_rep('S/F')
_rep('mS/uF')
_rep('mS/cm^2')
_rep('mS/mm^2')
_rep('L/F/s')
_rep('L/F/ms')
_rep('mA/cm^2')
# Capacitance
_rep('cm^2/uF')
_rep('uF/cm^2')
_rep('uF/mm^2')
_rep('uF*mol/C')
# Various
_rep('J/mol/K')
_rep('mJ/mol/K')
_rep('C/mol')
_rep('C/mmol')
_rep('J/K')  # Boltzmann
_rep('J*s')  # Planck
_rep('J*ms')



#
# Contains functions for the import and export of Myokit objects.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import os
import sys
import traceback
import warnings

import myokit

# Constants
DIR_FORMATS = os.path.join(myokit.DIR_MYOKIT, 'formats')

# Hidden lists
_IMPORTERS = None
_EXPORTERS = None
_EWRITERS = None


# Classes & methods
class Exporter:
    """
    Abstract base class for exporters.
    """
    def __init__(self):
        super().__init__()

    def post_export_info(self):
        """
        Optional method that returns a string containing information about this
        exporter, to be shown after the export is completed.
        """
        return ''

    def _test_writable_dir(self, path):
        """
        Ensures the given path is writable, or raises a
        :class:`myokit.ExportError` if it can't be used.
        """
        if os.path.exists(path):
            if not os.path.isdir(path):
                raise myokit.ExportError(
                    'Can\'t create output directory. A file exists at the'
                    ' specified location: ' + path)

        elif path != '' and not os.path.isdir(path):
            os.makedirs(path)

    def model(self, path, model):
        """
        Exports a :class:`myokit.Model`.

        The output will be stored in the **file** ``path``. A
        :class:`myokit.ExportError` will be raised if any errors occur.
        """
        raise NotImplementedError

    def runnable(self, path, model, protocol=None, *args):
        """
        Exports a :class:`myokit.Model` and optionally a
        :class:`myokit.Protocol` to something that can be run or compiled.

        The output will be stored in the **directory** ``path``. A
        :class:`myokit.ExportError` will be raised if any errors occur.
        """
        raise NotImplementedError

    def supports_model(self):
        """
        Returns ``True`` if this exporter supports model export.
        """
        return False

    def supports_runnable(self):
        """
        Returns ``True`` if this exporter supports export of a model and
        optional protocol to a runnable piece of code.
        """
        return False


def exporter(name):
    """
    Creates and returns an instance of the exporter specified by ``name``.
    """
    name = str(name)
    if _EXPORTERS is None:  # pragma: no cover
        _scan_for_internal_formats()
    try:
        return _EXPORTERS[name]()
    except KeyError:
        raise KeyError('Exporter not found: ' + name)


def exporters():
    """
    Returns a list of available exporters by name.
    """
    if _EXPORTERS is None:  # pragma: no cover
        _scan_for_internal_formats()
    return sorted(_EXPORTERS.keys())


class ExpressionWriter:
    """
    Base class for expression writers, that take myokit expressions as input
    and convert them to text or other formats.
    """
    def __init__(self):
        self._op_map = self._build_op_map()

    def _build_op_map(self):
        """
        Returns a mapping from myokit operators to lambda functions on expr.
        """
        return {
            myokit.Name: self._ex_name,
            myokit.Derivative: self._ex_derivative,
            myokit.PartialDerivative: self._ex_partial_derivative,
            myokit.InitialValue: self._ex_initial_value,
            myokit.Number: self._ex_number,
            myokit.PrefixPlus: self._ex_prefix_plus,
            myokit.PrefixMinus: self._ex_prefix_minus,
            myokit.Plus: self._ex_plus,
            myokit.Minus: self._ex_minus,
            myokit.Multiply: self._ex_multiply,
            myokit.Divide: self._ex_divide,
            myokit.Quotient: self._ex_quotient,
            myokit.Remainder: self._ex_remainder,
            myokit.Power: self._ex_power,
            myokit.Sqrt: self._ex_sqrt,
            myokit.Sin: self._ex_sin,
            myokit.Cos: self._ex_cos,
            myokit.Tan: self._ex_tan,
            myokit.ASin: self._ex_asin,
            myokit.ACos: self._ex_acos,
            myokit.ATan: self._ex_atan,
            myokit.Exp: self._ex_exp,
            myokit.Log: self._ex_log,
            myokit.Log10: self._ex_log10,
            myokit.Floor: self._ex_floor,
            myokit.Ceil: self._ex_ceil,
            myokit.Abs: self._ex_abs,
            myokit.Not: self._ex_not,
            myokit.Equal: self._ex_equal,
            myokit.NotEqual: self._ex_not_equal,
            myokit.More: self._ex_more,
            myokit.Less: self._ex_less,
            myokit.MoreEqual: self._ex_more_equal,
            myokit.LessEqual: self._ex_less_equal,
            myokit.And: self._ex_and,
            myokit.Or: self._ex_or,
            myokit.If: self._ex_if,
            myokit.Piecewise: self._ex_piecewise,
        }

    def eq(self, q):
        """
        Converts an equation to a string
        """
        return self.ex(q.lhs) + ' = ' + self.ex(q.rhs)

    def ex(self, e):
        """
        Converts an Expression to a string.
        """
        t = type(e)
        if t not in self._op_map:   # pragma: no cover
            raise ValueError('Unknown expression type: ' + str(t))
        return self._op_map[t](e)

    def _ex_name(self, e):
        raise NotImplementedError

    def _ex_derivative(self, e):
        raise NotImplementedError

    def _ex_partial_derivative(self, e):
        raise NotImplementedError

    def _ex_initial_value(self, e):
        raise NotImplementedError

    def _ex_number(self, e):
        raise NotImplementedError

    def _ex_prefix_plus(self, e):
        raise NotImplementedError

    def _ex_prefix_minus(self, e):
        raise NotImplementedError

    def _ex_plus(self, e):
        raise NotImplementedError

    def _ex_minus(self, e):
        raise NotImplementedError

    def _ex_multiply(self, e):
        raise NotImplementedError

    def _ex_divide(self, e):
        raise NotImplementedError

    def _ex_quotient(self, e):
        raise NotImplementedError

    def _ex_remainder(self, e):
        raise NotImplementedError

    def _ex_power(self, e):
        raise NotImplementedError

    def _ex_sqrt(self, e):
        raise NotImplementedError

    def _ex_sin(self, e):
        raise NotImplementedError

    def _ex_cos(self, e):
        raise NotImplementedError

    def _ex_tan(self, e):
        raise NotImplementedError

    def _ex_asin(self, e):
        raise NotImplementedError

    def _ex_acos(self, e):
        raise NotImplementedError

    def _ex_atan(self, e):
        raise NotImplementedError

    def _ex_exp(self, e):
        raise NotImplementedError

    def _ex_log(self, e):
        raise NotImplementedError

    def _ex_log10(self, e):
        raise NotImplementedError

    def _ex_floor(self, e):
        raise NotImplementedError

    def _ex_ceil(self, e):
        raise NotImplementedError

    def _ex_abs(self, e):
        raise NotImplementedError

    def _ex_not(self, e):
        raise NotImplementedError

    def _ex_equal(self, e):
        raise NotImplementedError

    def _ex_not_equal(self, e):
        raise NotImplementedError

    def _ex_more(self, e):
        raise NotImplementedError

    def _ex_less(self, e):
        raise NotImplementedError

    def _ex_more_equal(self, e):
        raise NotImplementedError

    def _ex_less_equal(self, e):
        raise NotImplementedError

    def _ex_and(self, e):
        raise NotImplementedError

    def _ex_or(self, e):
        raise NotImplementedError

    def _ex_if(self, e):
        raise NotImplementedError

    def _ex_piecewise(self, e):
        raise NotImplementedError

    def set_lhs_function(self, f):
        """
        Sets a naming function, will be called to get the variable name from a
         ``myokit.LhsExpression`` object.

        The argument ``f`` should be a function that takes an ``LhsExpression``
        as input and returns a string.
        """
        raise NotImplementedError


def ewriter(name):
    """
    Creates and returns an instance of the expression writer specified by
    ``name``.
    """
    name = str(name)
    if _EWRITERS is None:   # pragma: no cover
        _scan_for_internal_formats()
    try:
        return _EWRITERS[name]()
    except KeyError:
        raise KeyError('Expression writer not found: ' + name)


def ewriters():
    """
    Returns a list of available expression writers by name.
    """
    if _EWRITERS is None:   # pragma: no cover
        _scan_for_internal_formats()
    return sorted(_EWRITERS.keys())


class Importer:
    """
    Abstract base class for importers.
    """
    def __init__(self):
        super().__init__()

    def component(self, path, model):
        """
        Imports a component from the given ``path`` and adds it to the given
        model.

        The importer may pose restraints on the used model. For example, the
        model may be required to contain a variable labelled
        "membrane_potential". For details, check the documentation of the
        individual importers.

        The created :class:`myokit.Component` is returned. A
        :class:`myokit.ImportError` will be raised if any errors occur.
        """
        raise NotImplementedError

    def model(self, path):
        """
        Imports a model from the given ``path``.

        The created :class:`myokit.Model` is returned. A
        :class:`myokit.ImportError` will be raised if any errors occur.
        """
        raise NotImplementedError

    def protocol(self, path):
        """
        Imports a protocol from the given ``path``.

        The created :class:`myokit.Protocol` is returned. A
        :class:`myokit.ImportError` will be raised if any errors occur.
        """
        raise NotImplementedError

    def supports_component(self):
        """
        Returns a bool indicating if component import is supported.
        """
        return False

    def supports_model(self):
        """
        Returns a bool indicating if model import is supported.
        """
        return False

    def supports_protocol(self):
        """
        Returns a bool indicating if protocol import is supported.
        """
        return False


def importer(name):
    """
    Creates and returns an instance of the importer specified by ``name``.
    """
    name = str(name)
    if _IMPORTERS is None:  # pragma: no cover
        _scan_for_internal_formats()
    try:
        return _IMPORTERS[name]()
    except KeyError:
        raise KeyError('Importer not found: ' + name)


def importers():
    """
    Returns a list of available importers by name.
    """
    if _IMPORTERS is None:  # pragma: no cover
        _scan_for_internal_formats()
    return sorted(_IMPORTERS.keys())


class TemplatedRunnableExporter(Exporter):
    """
    *Abstract class, extends:* :class:`Exporter`

    Abstract base class for exporters that turn a model (and optionally a
    protocol) into a runnable chunk of code.
    """
    def __init__(self):
        super().__init__()

    def runnable(self, path, model, protocol=None, *args):
        """
        Exports a :class:`myokit.Model` and optionally a
        :class:`myokit.Protocol` to something that can be run or compiled.

        The output will be stored in the **directory** ``path``.
        """
        # Get and test path
        path = os.path.abspath(os.path.expanduser(path))
        path = myokit.tools.format_path(path)
        self._test_writable_dir(path)

        # Clone the model, allowing changes to be made during export
        model = model.clone()
        model.validate()

        # Ensure we have a protocol
        if protocol is None:
            protocol = myokit.default_protocol()

        # Build and check template path
        tpl_dir = self._dir(DIR_FORMATS)
        if not os.path.exists(tpl_dir):  # pragma: no cover
            # Cover pragma: If this happens it's a bug in the exporter
            msg = 'Template directory not found: ' + tpl_dir
            raise myokit.ExportError(msg)
        if not os.path.isdir(tpl_dir):  # pragma: no cover
            # Cover pragma: If this happens it's a bug in the exporter
            msg = 'Template path is not a directory:' + tpl_dir
            raise Myokit.ExportError(msg)

        # Render all templates
        tpl_vars = self._vars(model, protocol, *args)
        for tpl_name, out_name in self._dict().items():

            # Create any dirs embedded in output file path
            file_dir = os.path.split(out_name)[0]

            if file_dir:
                file_dir = os.path.join(path, file_dir)
                if os.path.exists(file_dir):
                    if not os.path.isdir(file_dir):
                        msg = 'Failed to create directory at: '
                        msg += myokit.tools.format_path(file_dir)
                        msg += ' A file or link with that name already exists.'
                        raise myokit.ExportError(msg)
                else:   # pragma: no cover
                    try:
                        os.makedirs(file_dir)
                    except IOError as e:
                        msg = 'Failed to create directory at: '
                        msg += myokit.tools.format_path(file_dir)
                        msg += ' IOError:' + str(e)
                        raise myokit.ExportError(msg)

            # Check if output file already exists
            out_name = os.path.join(path, out_name)
            if os.path.exists(out_name):
                if os.path.isdir(out_name):
                    msg = 'Directory exists at ' \
                          + myokit.tools.format_path(out_name)
                    raise myokit.ExportError(msg)

            # Check template file
            tpl_name = os.path.join(tpl_dir, tpl_name)
            if not os.path.exists(tpl_name):    # pragma: no cover
                # Cover pragma: If this happens it's a bug in the exporter
                msg = 'File not found: ' + myokit.tools.format_path(tpl_name)
                raise myokit.ExportError(msg)
            if not os.path.isfile(tpl_name):    # pragma: no cover
                # Cover pragma: If this happens it's a bug in the exporter
                msg = 'Directory found, expecting file at '
                msg += myokit.tools.format_path(tpl_name)
                raise myokit.ExportError(msg)

            # Render
            with open(out_name, 'w') as f:
                p = myokit.pype.TemplateEngine()
                p.set_output_stream(f)
                try:
                    p.process(tpl_name, tpl_vars)
                except Exception as e:      # pragma: no cover
                    warnings.warn(
                        'An error ocurred while processing the template at '
                        + myokit.tools.format_path(tpl_name))
                    warnings.warn(traceback.format_exc())
                    if isinstance(e, myokit.pype.PypeError):
                        # Pype error? Then add any error details
                        d = p.error_details()
                        if d:
                            warnings.warn(d)
                    raise myokit.ExportError(   # pragma: no cover
                        'An internal error ocurred while processing a'
                        ' template.')

    def supports_runnable(self):
        """
        Returns ``True`` if this exporter supports export of a model and
        optional protocol to a runnable piece of code.
        """
        return True

    def _dict(self):
        """
        Returns a dict (filename : template_file_name) containing all the
        templates used by this exporter.

        *This should be implemented by each subclass.*
        """
        raise NotImplementedError

    def _dir(self, formats_dir):
        """
        Returns the path to this exporter's data files (as a string).

        *This should be implemented by each subclass. The root directory all
        format extensions are stored in in passed in as ``formats_dir``.*
        """
        raise NotImplementedError

    def _vars(self, model, protocol):
        """
        Returns a dict containing all variables the templates will need.

        Will be called with the arguments `model` and `protocol`, followed by
        any extra arguments passed to :meth:`runnable`.

        *This should be implemented by each subclass.*
        """
        raise NotImplementedError


def _scan_for_internal_formats():
    """
    Scans for importers, exporters and expression writers.
    """
    global _IMPORTERS, _EXPORTERS, _EWRITERS
    if _IMPORTERS is None:
        _IMPORTERS = {}
    if _EXPORTERS is None:
        _EXPORTERS = {}
    if _EWRITERS is None:
        _EWRITERS = {}
    for fname in os.listdir(DIR_FORMATS):
        d = os.path.join(DIR_FORMATS, fname)
        if not os.path.isdir(d):
            continue

        # Only check modules
        f = os.path.join(d, '__init__.py')
        if not os.path.isfile(f):
            continue     # pragma: no cover

        # Dynamically load module
        name = 'myokit.formats.' + fname
        __import__(name)
        m = sys.modules[name]

        # Add importers, exporters and expression writers to global list
        try:
            x = m.importers()
        except AttributeError:
            x = {}
        for k, v in x.items():
            if k in _IMPORTERS:     # pragma: no cover
                raise Exception('Duplicate importer name: "' + str(k) + '".')
            _IMPORTERS[k] = v

        try:
            x = m.exporters()
        except AttributeError:
            x = {}
        for k, v in x.items():
            if k in _EXPORTERS:     # pragma: no cover
                raise Exception('Duplicate exporter name: "' + str(k) + '".')
            _EXPORTERS[k] = v

        try:
            x = m.ewriters()
        except AttributeError:
            x = {}
        for k, v in x.items():
            if k in _EWRITERS:     # pragma: no cover
                raise Exception(
                    'Duplicate expression writer name: "' + str(k) + '".')
            _EWRITERS[k] = v


def register_external_importer(name, importer_class):
    """
    Registers an external :class:`Importer` for use with Myokit.

    Arguments:

    ``name``
        A short descriptive string name.
    ``importer_class``
        The class to register (must be a :class:`myokit.Importer`).
        Importers can be unregistered by passing in ``None``.

    """
    if importer_class is None:
        if _IMPORTERS is not None and name in _IMPORTERS:
            del _IMPORTERS[name]
    else:
        if _IMPORTERS is None:  # pragma: no cover
            _scan_for_internal_formats()
        _IMPORTERS[name] = importer_class


def register_external_exporter(name, exporter_class):
    """
    Registers an external :class:`Exporter` for use with Myokit.

    Arguments:

    ``name``
        A short descriptive string name.
    ``exporter_class``
        The class to register (must be a :class:`myokit.Exporter`).
        Exporters can be unregistered by passing in ``None``.

    """
    if exporter_class is None:
        if _EXPORTERS is not None and name in _EXPORTERS:
            del _EXPORTERS[name]
    else:
        if _EXPORTERS is None:  # pragma: no cover
            _scan_for_internal_formats()
        _EXPORTERS[name] = exporter_class


def register_external_ewriter(name, ewriter_class):
    """
    Registers an external :class:`ExpressionWriter` for use with Myokit.

    Arguments:

    ``name``
        A short descriptive string name.
    ``ewriter_class``
        The class to register (must be a :class:`myokit.ExpressionWriter`).
        Expression writers can be unregistered by passing in ``None``.

    """
    if ewriter_class is None:
        if _EWRITERS is not None and name in _EWRITERS:
            del _EWRITERS[name]
    else:
        if _EWRITERS is None:  # pragma: no cover
            _scan_for_internal_formats()
        _EWRITERS[name] = ewriter_class


class SweepSource:
    """
    Interface for classes that provide time-series data organised into *sweeps*
    (or *records* or *episodes*) and *channels* (or *traces*).

    The :class:`SweepSource` interface defines methods to get the number of
    sweeps and channels, the names and units of the channels, and the data
    stored in channels either as numpy arrays or in a :class:`myokit.DataLog`.

    Each sweep contains the same number of channels, and each channel is
    represented as a 1d array. In most cases these arrays have the same length
    for every sweep, but whether this is the case for the current source can be
    tested with :meth:`equal_length_sweeps`.

    Data can be retrieved sweep by sweep::

        (time[0], sweep[0]),        n_t data points, n_c channels of n_t points
        (time[1], sweep[1]),
        (time[2], sweep[2]),
        ...
        (time[n_s], sweep[n_s])

    This allows plotting in the common "overlaid" fashion, i.e. plotting every
    ``sweep[i] against ``time[0]``.

    For other types of analysis (e.g. parameter estimation) the data can also
    be returned as single time series::

        time                            sum(n_t[i]) data points
        sweep[0] + sweep[1] + ...       n_c channels, with sum(n_t[i]) points

    Some formats can also contain information from which D/A output signals can
    be reconstructed. These can be accessed using the :meth:`da`.

    """
    def channel(self, channel_id, join_sweeps=False):
        """
        Returns the data for a single channel, identified by integer or string
        ``channel_id``.

        With ``join_sweeps=False``, the data is returned as a tuple
        ``(times, sweeps)`` where ``times`` and ``sweeps`` are 2d numpy arrays
        indexed so that ``times[i][j]`` is the ``j``-th time point for sweep
        ``i``.

        If ``join_sweeps=True`` the sweeps are joined together, and a tuple
        ``(times, values)`` is returned ``times`` and ``values`` are 1d arrays.
        """
        raise NotImplementedError

    def channel_count(self):
        """ Returns the number of channels. """
        raise NotImplementedError

    def channel_names(self, index=None):
        """
        Returns the names of all channels or the name of a specific channel
        ``index``.
        """
        raise NotImplementedError

    def channel_units(self, index=None):
        """
        Returns the units (as :class:`myokit.Unit`) of all channels or the
        units of a specific channel ``index``.
        """
        raise NotImplementedError

    def da(self, output_id, join_sweeps=False):
        """ Like :meth:`channel`, but returns reconstructed D/A signals. """
        raise NotImplementedError

    def da_count(self):
        """
        Returns the available number of reconstructed D/A output channels.

        This should return 0 if D/A channels are not supported.
        """
        raise NotImplementedError

    def da_names(self, index=None):
        """
        Returns the names of all reconstructed D/A output channels or the name
        of a specific output channel ``index``.

        This will raise a ``NotImplementedError`` if D/A channels are not
        supported.
        """
        raise NotImplementedError

    def da_protocol(self, output_id=None, tu='ms', vu='mV', cu='pA',
                    n_digits=9):
        """
        Returns a :class:`myokit.Protocol` representation of the D/A output
        channel specified by name or integer ``output_id``.

        If no explicit output channel is set, a guess will be made.

        Time units will be converted to ``tu``, voltage units to ``vu``, and
        current units to ``cu``. Other unit types will be left unchanged. Units
        may be specified as :class:`myokit.Unit` objects or strings.

        By default, floating point numbers in the protocol will be rounded to
        9 digits after the decimal point. This can be customised using the
        argument ``n_digits``.

        If a D/A output cannot be converted to a :class:`myokit.Protocol`, a
        ``ValueError`` is raised. A ``NotImplementedError`` is raised if D/A
        channels are not supported at all.
        """
        raise NotImplementedError

    def da_units(self, index=None):
        """
        Returns the units (as :class:`myokit.Unit`) of all reconstructed D/A
        output channels or the units of a specific output channel ``index``.

        This will raise a ``NotImplementedError`` if D/A channels are not
        supported.
        """
        raise NotImplementedError

    def equal_length_sweeps(self):
        """
        Returns ``True`` only if each sweep in this source has the same length.
        """
        raise NotImplementedError

    def log(self, join_sweeps=False, use_names=False, include_da=True):
        """
        Returns a :class:`myokit.DataLog` containing the data from all
        channels.

        The log will have a single entry ``time`` corresponding to the time of
        the first sweep if ``join_sweeps`` is ``False``, or the time of all
        points when ``join_sweeps`` is ``True``.

        Names will have a systematic form ``i_sweep.i_channel.label``, for
        example ``0.1.channel`` for sweep 0 of recorded channel 1, or
        ``3.0.da`` for sweep 3 of reconstructed D/A output 0. These can also be
        accessed using the syntax ``log['channel', 1, 0]`` and
        ``log['da', 0, 3]``.

        To obtain a log with the user-specified names from the source instead,
        set ``use_names`` to ``True``. This will result in names such as
        ``0.IN 1`` or ``3.Cmd 0``.

        To exclude D/A signal reconstructions, set ``include_da`` to ``False``.

        A call with ``join_sweeps=False`` on a source where
        :meth:`equal_length_sweeps()` returns ``False`` will raise a
        ``ValueError``.
        """
        raise NotImplementedError

    def meta_str(self, verbose=False):
        """
        Optional method that returns a multi-line string with unstructured meta
        data about the source and its contents.

        Will return ``None`` if no such string is available.
        """
        return None  # pragma: no cover

    def sweep_count(self):
        """
        Returns the number of sweeps.

        Note that a source with zero recorded channels may still report a
        non-zero number of sweeps if it can provide D/A outputs.
        """
        raise NotImplementedError

    def time_unit(self):
        """ Returns the time unit used in this source. """
        raise NotImplementedError




#
# Converts SBML to Myokit expressions, using an ElementTree implementation.
#
# Only partial SBML support (based on SBML level 3 version 2) is provided.
# The SBML file format specifications can be found here
# http://sbml.org/Documents/Specifications.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import re
import warnings

from lxml import etree

import myokit
import myokit.units
import myokit.formats
import myokit.formats.html
import myokit.formats.mathml
import myokit.formats.sbml
import myokit.formats.xml


# Namespaces
NS_MATHML = 'http://www.w3.org/1998/Math/MathML'
NS_SBML_3_2 = 'http://www.sbml.org/sbml/level3/version2/core'


# Regex for id checking
_re_id = re.compile(r'^[a-zA-Z_]+[a-zA-Z0-9_]*$')


def _parse_mathml(expr, name_generator):
    """
    Parses the MathML expression ``expr`` and wraps any errors in an SBMLError.
    """
    def name(x, y):
        try:
            return name_generator(x, y)
        except KeyError:
            raise myokit.formats.sbml.SBMLError(
                'Unknown or inaccessible symbol in MathML: "' + str(x) + '".')

    try:
        return myokit.formats.mathml.parse_mathml_etree(
            expr, name, lambda x, y: myokit.Number(x))
    except myokit.formats.mathml.MathMLError as e:
        raise myokit.formats.sbml.SBMLError(
            'Unable to parse MathML: ' + str(e))

    # Note: In any place we parse mathematics, we could also check that
    # there's nothing after the mathematics. E.g.
    #   <math xmlns="..">
    #    <cn>1</cn>
    #    <cn>2</cn>
    #   </math>
    # is currently allowed because only the first statement is read.
    # But for now we're being lenient.


class SBMLParsingError(myokit.ImportError):
    """
    Thrown if an error occurs when parsing SBML.

    The argument ``element`` can be used to pass in an element that caused the
    error.
    """
    def __init__(self, message, element=None):
        if element is not None:
            try:
                line = str(element.sourceline)
                message = 'Error on line ' + line + '. ' + message
            except AttributeError:  # pragma: no cover
                pass
        super().__init__(message)


class SBMLParser:
    """
    Parses SBML models, creating an SBML Model structure that can be converted
    to a :class:`myokit.Model` object.

    Support notes:

    - SBML older than Level 2 version 1 is unlikely to work.
    - Algebraic rules are not supported.
    - Constraints are not supported.
    - Events are not supported.
    - Function definitions are not supported.
    - Local parameters in reactions are not supported.
    - Units "celsius" are not supported.
    - Since SBML does not support units in equations, variable units will be
      set, but RHS units will not.

    """

    def _check_required_attributes(self, element, *attributes):
        """
        Checks that an ``elementtree`` element has all the required attributes.
        """
        for attribute in attributes:
            if element.get(attribute) is None:
                raise SBMLParsingError(
                    'Element ' + self._tag(element) + ' is missing required'
                    ' attribute "' + str(attribute) + '".', element)

    def parse_file(self, path):
        """
        Parses the SBML file at ``path`` and returns a :class:`myokit.Model`.
        """
        # Read file
        try:
            parser = etree.XMLParser(remove_comments=True)
            tree = etree.parse(path, parser=parser)
        except Exception as e:
            raise SBMLParsingError('Unable to parse XML: ' + str(e))

        # Parse content
        return self.parse(tree.getroot())

    def parse_string(self, text):
        """
        Parses the SBML XML in the string ``text`` and returns a
        :class:`myokit.Model`.
        """
        # Read string
        try:
            root = etree.fromstring(text)
        except Exception as e:
            raise SBMLParsingError('Unable to parse XML: ' + str(e))

        # Parse content
        return self.parse(root)

    def parse(self, element):
        """
        Parses an SBML document rooted in the given ``ElementTree`` element and
        returns a :class:`myokit.Model`.
        """
        # Supported namespaces
        # Other namespaces are allowed, but might not work.
        supported = (
            NS_SBML_3_2,
        )

        # Check whether document declares a supported namespace
        self._ns = myokit.formats.xml.split(element.tag)[0]
        try:
            if self._ns not in supported:
                warnings.warn(
                    'Unknown SBML namespace ' + str(self._ns) + '. This'
                    ' version of SBML may not be supported.')

            # Get model and parse
            model = element.find(self._path('model'))
            if model is None:
                raise SBMLParsingError('Model element not found.', element)
            return self._parse_model(model)

        finally:
            # Remove all references to temporary state
            del self._ns

    def _parse_compartment(self, element, model):
        """Parses a ``compartment`` element."""

        # Check required attributes
        self._check_required_attributes(element, 'id')
        # The attribute 'constant' is also required in level 3, but optional in
        # level 2 (default true). If constant is false then the compartment's
        # size can be changed.

        # Note: Any place we check required attributes, we could also check for
        # allowed attributes (e.g. errors for unrecognised ones). Bit tricky
        # for multiple SBML versions. For now we're being lenient.

        sid = element.get('id')
        try:
            # Create compartment
            compartment = model.add_compartment(sid)

            # Get units, if given
            units = element.get('units')
            if units is not None:
                try:
                    units = model.unit(units)
                except myokit.formats.sbml.SBMLError:
                    raise SBMLParsingError(
                        'Unknown units "' + units + '".', element)
                compartment.set_size_units(units)

            # Get spatial dimensions, if given
            dim = element.get('spatialDimensions')
            if dim is not None:
                try:
                    dim = float(dim)
                except ValueError:
                    raise SBMLParsingError(
                        'Unable to convert spatial dimensions value "'
                        + str(dim) + '" to float.', element)
                compartment.set_spatial_dimensions(dim)

            # Get compartment size, if set
            size = element.get('size')
            if size is not None:
                try:
                    size = float(size)
                except ValueError:
                    raise SBMLParsingError(
                        'Unable to convert size "' + str(size) + '" to float.',
                        element)

                # Ignore units in RHS
                compartment.set_initial_value(myokit.Number(size))

        except myokit.formats.sbml.SBMLError as e:
            raise SBMLParsingError(
                'Unable to parse compartment "' + sid + '": ' + str(e),
                element)

    def _parse_initial_assignment(self, element, model):
        """Parses an initial assignment."""

        # Check required attributes
        self._check_required_attributes(element, 'symbol')

        # Check that symbol refers to an assignable
        var = element.get('symbol')
        try:
            var = model.assignable(var)
        except KeyError:
            raise SBMLParsingError(
                'Unable to parse rule: The given SId "' + str(var) + '" does'
                ' not refer to a Compartment, Species, SpeciesReference, or'
                ' Parameter in this model.')

        try:
            # Find mathematics
            expr = element.find('{' + NS_MATHML + '}math')
            if expr is None:
                # This is allowed, so that SBML extensions can do mysterious
                # things.
                warnings.warn(
                    'Initial assignment does not define any mathematics')
                return

            # Parse rule and apply
            var.set_initial_value(_parse_mathml(
                expr,
                lambda x, y: myokit.Name(model.assignable_or_csymbol(x)),
            ))

        except myokit.formats.sbml.SBMLError as e:
            raise SBMLParsingError(
                'Unable to parse initial assignment: ' + str(e), element)

    def _parse_kinetic_law(self, element, model, reaction):
        """
        Parses the ``kinetic_law`` element used in reactions and returns the
        resulting :class:`myokit.Expression``.

        Returns ``None`` if no expression is found.
        """

        # Local parameters are not supported
        x = element.find(self._path('listOfLocalParameters', 'localParameter'))
        if x is not None:
            raise SBMLParsingError('Local parameters are not supported.', x)

        # Kinetic law is not allowed use species unless they're defined in the
        # reaction.
        def name(x, y):
            obj = model.assignable_or_csymbol(x)
            if isinstance(obj, myokit.formats.sbml.Species):
                obj = reaction.species(x)
            return myokit.Name(obj)

        # Find and parse expression
        expr = element.find('{' + NS_MATHML + '}math')
        if expr is None:
            return None
        try:
            return _parse_mathml(expr, name)
        except myokit.formats.sbml.SBMLError as e:
            raise SBMLParsingError(
                'Unable to parse kinetic law: ' + str(e), element)

    def _parse_model(self, element):
        """Parses a ``model`` element in an SBML document."""

        # Retrieve or create a model name.
        # SBML Models can have an optional name attribute (user friendly name)
        # or an optional id attribute (not necessarily user friendly) or can
        # have no name at all.
        name = element.get('name')
        if name is None:
            name = element.get('id')
        if name is None:
            name = 'Imported SBML model'

        # Create SBML model
        model = myokit.formats.sbml.Model(name)

        # Algebraic rules are not supported
        x = element.find(self._path('listOfRules', 'algebraicRule'))
        if x is not None:
            raise SBMLParsingError(
                'Algebraic rules are not supported.', x)

        # Constraints are not supported, but will be ignored
        x = element.find(self._path('listOfConstraints', 'constraint'))
        if x is not None:
            warnings.warn('Ignoring SBML constraints.')

        # Events are not supported, but will be ignored
        x = element.find(self._path('listOfEvents', 'event'))
        if x is not None:
            warnings.warn('Ignoring SBML events.')

        # Function definitions are not supported.
        x = element.find(self._path(
            'listOfFunctionDefinitions', 'functionDefinition'))
        if x is not None:
            raise SBMLParsingError(
                'Function definitions are not supported.', x)

        # Parse notes element (converts HTML to plain text)
        notes = element.find(self._path('notes'))
        if notes is not None:
            self._parse_notes(notes, model)

        # Parse unit definitions
        for unit in element.findall(self._path(
                'listOfUnitDefinitions', 'unitDefinition')):
            self._parse_unit_definition(unit, model)

        # Parse model unit attributes
        try:
            # Default units for compartment sizes
            units = element.get('lengthUnits')
            if units is not None:
                model.set_length_units(model.unit(units))
            units = element.get('areaUnits')
            if units is not None:
                model.set_area_units(model.unit(units))
            units = element.get('volumeUnits')
            if units is not None:
                model.set_volume_units(model.unit(units))

            # Default units for species amounts (not concentrations)
            units = element.get('substanceUnits')
            if units is not None:
                model.set_substance_units(model.unit(units))

            # Default units for reactions and rates.
            # The KineticLaw equations are in units of extentUnits / timeUnits.
            units = element.get('extentUnits')
            if units is not None:
                model.set_extent_units(model.unit(units))
            units = element.get('timeUnits')
            if units is not None:
                model.set_time_units(model.unit(units))

        except myokit.formats.sbml.SBMLError as e:
            raise SBMLParsingError(
                'Error parsing model element: ' + str(e), element)

        # Parse parameters
        for parameter in element.findall(self._path(
                'listOfParameters', 'parameter')):
            self._parse_parameter(parameter, model)

        # Set global conversion factor, if provided (must refer to a parameter)
        factor = element.get('conversionFactor')
        if factor is not None:
            try:
                factor = model.parameter(factor)
            except KeyError:
                raise SBMLParsingError(
                    'Model conversion factor "' + str(factor) + '" does not'
                    ' refer to a parameter SId.', element)
            model.set_conversion_factor(factor)

        # Parse compartments
        for compartment in element.findall(self._path(
                'listOfCompartments', 'compartment')):
            self._parse_compartment(compartment, model)

        # Parse species
        for species in element.findall(self._path('listOfSpecies', 'species')):
            self._parse_species(species, model)

        # Parse reactions
        for reaction in element.findall(self._path(
                'listOfReactions', 'reaction')):
            self._parse_reaction(reaction, model)

        # Parse initial assignments
        for assignment in element.findall(self._path(
                'listOfInitialAssignments', 'initialAssignment')):
            self._parse_initial_assignment(assignment, model)

        # Parse assignment rules
        for rule in element.findall(self._path(
                'listOfRules', 'assignmentRule')):
            self._parse_rule(rule, model, False)

        # Parse rate rules
        for rule in element.findall(self._path('listOfRules', 'rateRule')):
            self._parse_rule(rule, model, True)

        return model

    def _parse_notes(self, element, model):
        """Parses a model's ``notes`` element, converting it to plain text."""

        notes = etree.tostring(element).decode()
        notes = myokit.formats.html.html2ascii(notes, width=75)
        if notes:
            model.set_notes(notes)

    def _parse_parameter(self, element, model):
        """Parses a ``parameter`` element."""

        # Check required attributes
        self._check_required_attributes(element, 'id')
        # The attribute 'constant' is also required in level 3, but optional in
        # level 2 (default false)

        # Create
        try:
            parameter = model.add_parameter(element.get('id'))

            # Get optional units
            unit = element.get('units')
            if unit is not None:
                try:
                    unit = model.unit(unit)
                except myokit.formats.sbml.SBMLError:
                    raise SBMLParsingError(
                        'Unknown units "' + unit + '".', element)
                parameter.set_units(unit)

            # Get optional initial value
            value = element.get('value')
            if value is not None:
                try:
                    value = float(value)
                except ValueError:
                    raise SBMLParsingError(
                        'Unable to convert parameter value "' + str(value)
                        + '" to float.', element)

                # Ignore unit
                parameter.set_initial_value(myokit.Number(value))

        except myokit.formats.sbml.SBMLError as e:
            raise SBMLParsingError(
                'Unable to parse ' + self._tag(element) + ': ' + str(e),
                element)

    def _parse_reaction(self, element, model):
        """Parses a ``reaction`` element."""

        # Check required attributes
        self._check_required_attributes(element, 'id')
        # The attribute 'reversible' is also required, but doesn't affect the
        # maths, so we ignore it here.
        # The attribute 'compartment' is not required, and doesn't affect the
        # maths, so we ignore it here.

        # The 'fast' attribute was removed in L3V2 and is not supported here
        if element.get('fast') == 'true':
            raise SBMLParsingError(
                'Reactions with fast="true" are not supported.', element)

        # Create
        try:
            reaction = model.add_reaction(element.get('id'))
        except myokit.formats.sbml.SBMLError as e:
            raise SBMLParsingError(
                'Unable to parse ' + self._tag(element) + ': ' + str(e),
                element)

        # Add reactants, products, modifiers
        have_reactant_or_product = False
        for x in element.findall(self._path(
                'listOfReactants', 'speciesReference')):
            self._parse_species_reference(
                x, model, reaction,
                myokit.formats.sbml.SpeciesReference.REACTANT)
            have_reactant_or_product = True

        for x in element.findall(self._path(
                'listOfProducts', 'speciesReference')):
            self._parse_species_reference(
                x, model, reaction,
                myokit.formats.sbml.SpeciesReference.PRODUCT)
            have_reactant_or_product = True

        for x in element.findall(self._path(
                'listOfModifiers', 'modifierSpeciesReference')):
            self._parse_species_reference(
                x, model, reaction,
                myokit.formats.sbml.SpeciesReference.MODIFIER)

        # Raise error if neither reactants nor products is populated
        if not have_reactant_or_product:
            raise SBMLParsingError(
                'Reaction must have at least one reactant or product.',
                element)

        # Parse kinetic law
        kinetic_law = element.find(self._path('kineticLaw'))
        if kinetic_law is not None:
            kinetic_law = self._parse_kinetic_law(kinetic_law, model, reaction)
        if kinetic_law is None:
            warnings.warn(
                'No kinetic law set for reaction "' + reaction.sid() + '".')
        else:
            reaction.set_kinetic_law(kinetic_law)

    def _parse_rule(self, element, model, rate=False):
        """Parses an assignment or rate rule."""

        # Check required attributes
        self._check_required_attributes(element, 'variable')

        # Check that variable refers to an assignable
        var = element.get('variable')
        try:
            var = model.assignable(var)
        except KeyError:
            raise SBMLParsingError(
                'Unable to parse rule: The given SId "' + str(var) + '" does'
                ' not refer to a Compartment, Species, SpeciesReference, or'
                ' Parameter in this model.')

        try:
            # Check that this assignable can be changed with assignment rules
            if isinstance(var, myokit.formats.sbml.Species):
                if not var.is_boundary():
                    raise SBMLParsingError(
                        'Assignment or rate rule set for species that is'
                        ' created or destroyed in a reaction.', element)
            # TODO: Could also gather 'constant' attributes for all, and check
            # here. But maybe we're not all that interested in validation...

            # Find mathematics
            expr = element.find('{' + NS_MATHML + '}math')
            if expr is None:
                # This is allowed, so that SBML extensions can do mysterious
                # things.
                warnings.warn('Rule does not define any mathematics')
                return

            # Parse rule and apply
            var.set_value(_parse_mathml(
                expr,
                lambda x, y: myokit.Name(model.assignable_or_csymbol(x)),
            ), rate)

        except myokit.formats.sbml.SBMLError as e:
            raise SBMLParsingError('Unable to parse rule: ' + str(e), element)

    def _parse_species(self, element, model):
        """
        Adds species to references compartment in model.
        """
        # Check required attributes (note: constant is also required in L3)
        self._check_required_attributes(element, 'id', 'compartment')
        # The attributes 'hasOnlySubstaneUnits', 'boundaryCondition', and
        # 'constant' are required in level 3, but optional in level 2 (default
        # is false for all three).

        # Check if it's an amount or a concentration
        is_amount = element.get('hasOnlySubstanceUnits', 'false') == 'true'

        # Check if constant, and if at a reaction boundary
        is_constant = element.get('constant', 'false') == 'true'
        is_boundary = element.get('boundaryCondition', 'false') == 'true'

        # Note: In lines like the above we could raise an error if the value
        # isn't 'true' or false', but for now we're being lenient.

        # Get compartment
        compartment = element.get('compartment')
        try:
            compartment = model.compartment(compartment)
        except KeyError:
            raise SBMLParsingError(
                'Unknown compartment "' + compartment + '"', element)

        # Create
        try:
            species = model.add_species(
                compartment,
                element.get('id'),
                is_amount,
                is_constant,
                is_boundary)

            # Set units, if provided
            units = element.get('substanceUnits')
            if units is not None:
                try:
                    units = model.unit(units)
                except myokit.formats.sbml.SBMLError:
                    raise SBMLParsingError(
                        'Unknown units "' + units + '"', element)

                # Set substance units, not concentration units
                species.set_substance_units(units)

            # Check that not both initialAmount and initialConcentration are
            # provided
            if (element.get('initialAmount') is not None) and (
                    element.get('initialConcentration') is not None):
                raise SBMLParsingError(
                    'Species cannot set both an initialAmount and an'
                    ' initialConcentration.')

            # Get initial amount
            value = element.get('initialAmount')

            # Indicate whether units of initial value are in amount
            value_in_amount = True

            # If initial amount is not provided, get initial concentration
            if value is None:
                value = element.get('initialConcentration')

                # Indicate whether units of initial value are correct
                value_in_amount = False

            # Set initial value
            if value is not None:
                try:
                    value = float(value)
                except ValueError:
                    raise SBMLParsingError(
                        'Unable to convert initial species value to float "'
                        + str(value) + '".', element)

                # Ignore units in equations
                species.set_initial_value(
                    myokit.Number(value), value_in_amount)

            # Set conversion factor if provided (must refer to a parameter)
            factor = element.get('conversionFactor')
            if factor is not None:
                try:
                    factor = model.parameter(factor)
                except KeyError:
                    raise SBMLParsingError(
                        'Unknown parameter "' + str(factor) + '" set as'
                        ' conversion factor.', element)
                species.set_conversion_factor(factor)

        except myokit.formats.sbml.SBMLError as e:
            raise SBMLParsingError(
                'Unable to parse ' + self._tag(element) + ': ' + str(e),
                element)

    def _parse_species_reference(self, element, model, reaction, ref_type):
        """Parses a ``speciesReference`` element inside a reaction."""

        # Check required attributes
        # Note: Level 3 also has a required attribute 'constant'
        self._check_required_attributes(element, 'species')

        # Find species
        species = element.get('species')
        try:
            species = model.species(species)
        except KeyError:
            raise SBMLParsingError(
                'Reference to unknown species "' + species + '"', element)

        # Get optional SId
        sid = element.get('id')

        try:
            # Create
            if ref_type == myokit.formats.sbml.SpeciesReference.REACTANT:
                reference = reaction.add_reactant(species, sid)
            elif ref_type == myokit.formats.sbml.SpeciesReference.PRODUCT:
                reference = reaction.add_product(species, sid)
            else:
                # Modifier references require no further parsing
                return reaction.add_modifier(species, sid)

            # Get optional stoichiometry
            stoichiometry = element.get('stoichiometry')
            if stoichiometry is not None:
                try:
                    stoichiometry = float(stoichiometry)
                except ValueError:
                    raise SBMLParsingError(
                        'Unable to convert stoichiometry value "'
                        + str(stoichiometry) + '" to float.', element)
                reference.set_initial_value(myokit.Number(stoichiometry))

        except myokit.formats.sbml.SBMLError as e:
            raise SBMLParsingError(
                'Unable to parse species reference: ' + str(e), element)

        return reference

    def _parse_unit_definition(self, element, model):
        """Parses a ``unitDefinition`` element."""

        # Check required attributes
        self._check_required_attributes(element, 'id')

        # Parse units and combine
        myokit_unit = myokit.units.dimensionless
        for unit in element.findall(self._path('listOfUnits', 'unit')):
            myokit_unit *= self._parse_unit(unit, model)

        # Store the unit
        sid = element.get('id')
        try:
            model.add_unit(sid, myokit_unit)
        except myokit.formats.sbml.SBMLError as e:
            raise SBMLParsingError(
                'Unable to parse unit definition for "' + sid + '": ' + str(e),
                element)

    def _parse_unit(self, element, model):
        """Parses a ``unit`` element and returns a :class:`myokit.Unit`."""

        # Check required attributes
        self._check_required_attributes(element, 'kind')
        # The attributes 'multiplier', 'scale', and 'exponent' are required in
        # level 3, but are optional in level 2.

        # Get base unit (must be a predefined type)
        try:
            unit = model.base_unit(element.get('kind'))
        except myokit.formats.sbml.SBMLError as e:
            raise SBMLParsingError(
                'Unable to parse unit kind: ' + str(e), element)

        # Parse remaining parts and return
        try:
            unit *= float(element.get('multiplier', default=1))
            unit *= 10 ** float(element.get('scale', default=0))
            unit **= float(element.get('exponent', default=1))
        except ValueError:
            raise SBMLParsingError('Unable to parse unit attributes', element)

        return unit

    def _path(self, *tags):
        """
        Returns a string created by prepending the namespace to each tag and
        adding forward slashes to separate.

        If a tag starts with a forward slash or period, no namespace will be
        prepended.
        """
        treated = []
        for tag in tags:
            if tag[:1] not in './':
                tag = '{' + self._ns + '}' + tag
            tag = tag.rstrip('/')
            treated.append(tag)
        return '/'.join(treated)

    def _tag(self, element):
        """
        Returns an element's name, but changes the syntax from ``{...}tag`` to
        ``sbml:tag`` for SBML elements.
        """
        _, el = myokit.formats.xml.split(element.tag)

        # Replace known namespaces
        tag = 'sbml:' + el

        # Add id if known
        sid = element.get('id', element.get('unitsid'))
        if sid is not None:
            tag += '[@id=' + str(sid) + ']'

        return tag

